#    Copyright 2019 George Melikov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import unittest

from restalchemy.storage.sql import engines
from restalchemy.storage.sql import sessions

from restalchemy.common import contexts


class SomeError(Exception):
    pass


@mock.patch("restalchemy.storage.sql.engines.engine_factory")
class TestContext(unittest.TestCase):
    def _configure_mocks(self, engine_factory_mock):
        self._engine = mock.Mock(spec=engines.MySQLEngine)
        self._storage = mock.Mock(spec=sessions.SessionThreadStorage)
        self._session = mock.Mock(spec=sessions.MySQLSession)
        # Configure mocks
        engine_factory_mock.configure_mock(
            **{
                "get_engine.return_value": self._engine,
            }
        )
        self._engine.configure_mock(
            **{
                "get_session.return_value": self._session,
                "get_session_storage.return_value": self._storage,
            }
        )
        self._storage.configure_mock(
            **{
                "get_session.return_value": self._session,
            }
        )

    def test_context_manager_no_exceptions(self, engine_factory_mock):
        self._configure_mocks(engine_factory_mock)
        context = contexts.Context()

        with context.session_manager():
            pass

        self._session.commit.assert_called_once()
        self._session.close.assert_called_once()
        self._storage.remove_session.assert_called_once()

    def test_context_manager_service_error(self, engine_factory_mock):
        self._configure_mocks(engine_factory_mock)
        context = contexts.Context()

        with self.assertRaises(SomeError):
            with context.session_manager():
                raise SomeError()

        self._session.commit.assert_not_called()
        self._session.rollback.assert_called_once()
        self._session.close.assert_called_once()
        self._storage.remove_session.assert_called_once()

    def test_context_manager_close_exception(self, engine_factory_mock):
        self._configure_mocks(engine_factory_mock)
        self._session.close.side_effect = SomeError
        context = contexts.Context()

        with context.session_manager():
            pass

        self._session.commit.assert_called_once()
        self._session.close.assert_called_once()
        self._storage.remove_session.assert_called_once()

    def test_context_manager_commit_exception(self, engine_factory_mock):
        self._configure_mocks(engine_factory_mock)
        self._session.commit.side_effect = SomeError
        context = contexts.Context()

        with self.assertRaises(SomeError):
            with context.session_manager():
                pass

        self._session.commit.assert_called_once()
        self._session.rollback.assert_called_once()
        self._session.close.assert_called_once()
        self._storage.remove_session.assert_called_once()

    def test_context_manager_commit_and_rollback_exception(self, engine_factory_mock):
        self._configure_mocks(engine_factory_mock)
        self._session.configure_mock(
            **{
                "commit.side_effect": SomeError,
                "rollback.side_effect": SomeError,
            }
        )
        context = contexts.Context()

        with self.assertRaises(SomeError):
            with context.session_manager():
                pass

        self._session.commit.assert_called_once()
        self._session.rollback.assert_called_once()
        self._session.close.assert_called_once()
        self._storage.remove_session.assert_called_once()



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# TODO(Eugene Frolov): Rewrite tests
import orjson

import mock
import webob

from restalchemy.api import field_permissions
from restalchemy.api import packers
from restalchemy.api import resources
from restalchemy.common import exceptions
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.tests.unit import base


class FakeModel(models.ModelWithUUID):
    field1 = properties.property(types.Integer(), required=False)
    field2 = properties.property(types.Integer())
    field3 = properties.property(types.Integer())
    field4 = properties.property(types.Integer(), required=True)


class TestData(object):
    uuid = None
    field1 = None
    field2 = 2
    field3 = 3
    field4 = 4


class BasePackerTestCase(base.BaseTestCase):
    def setUp(self):
        super(BasePackerTestCase, self).setUp()
        self._test_instance = packers.BaseResourcePacker(
            resources.ResourceByRAModel(FakeModel), mock.Mock()
        )

    def tearDown(self):
        super(BasePackerTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._test_instance

    def test_none_field_value(self):
        test_data = {"field1": None}

        result = self._test_instance.unpack(test_data)

        self.assertDictEqual(result, test_data)


class PackerFieldPermissionsHiddenTestCase(base.BaseTestCase):
    def setUp(self):
        req = mock.Mock()
        req.context.roles = ["owner"]

        super(PackerFieldPermissionsHiddenTestCase, self).setUp()
        self._test_resource_packer = packers.BaseResourcePacker(
            resources.ResourceByRAModel(
                FakeModel,
                fields_permissions=field_permissions.FieldsPermissionsByRole(
                    default=field_permissions.UniversalPermissions(
                        permission=field_permissions.Permissions.HIDDEN
                    )
                ),
            ),
            req,
        )

    def tearDown(self):
        super(PackerFieldPermissionsHiddenTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._test_resource_packer

    def test_pack(self):
        new_data = TestData()
        expected_data = {}

        result = self._test_resource_packer.pack(new_data)
        self.assertDictEqual(result, expected_data)

    def test_unpack(self):
        new_data = {"field2": 2}

        with self.assertRaises(exceptions.FieldPermissionError) as context:
            self._test_resource_packer.unpack(new_data)

        self.assertEqual("Permission denied for field field2.", str(context.exception))
        self.assertEqual(context.exception.code, 403)


class PackerFieldPermissionsNonDefaultHiddenTestCase(base.BaseTestCase):
    def setUp(self):
        req = mock.Mock()
        req.context.roles = ["owner"]

        super().setUp()
        self._test_resource_packer = packers.BaseResourcePacker(
            resources.ResourceByRAModel(
                FakeModel,
                fields_permissions=field_permissions.FieldsPermissionsByRole(
                    default=field_permissions.UniversalPermissions(
                        permission=field_permissions.Permissions.RW
                    ),
                    owner=field_permissions.UniversalPermissions(
                        permission=field_permissions.Permissions.HIDDEN
                    ),
                ),
            ),
            req,
        )

    def tearDown(self):
        super().tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._test_resource_packer

    def test_pack(self):
        new_data = TestData()
        expected_data = {}

        result = self._test_resource_packer.pack(new_data)
        self.assertDictEqual(result, expected_data)

    def test_unpack(self):
        new_data = {"field2": 2}

        with self.assertRaises(exceptions.FieldPermissionError) as context:
            self._test_resource_packer.unpack(new_data)

        self.assertEqual("Permission denied for field field2.", str(context.exception))
        self.assertEqual(context.exception.code, 403)


class PackerFieldPermissionsRWTestCase(base.BaseTestCase):
    def setUp(self):
        req = mock.Mock()
        req.context.roles = ["owner"]

        super(PackerFieldPermissionsRWTestCase, self).setUp()
        self._test_resource_packer = packers.BaseResourcePacker(
            resources.ResourceByRAModel(
                FakeModel,
                fields_permissions=field_permissions.FieldsPermissionsByRole(
                    default=field_permissions.UniversalPermissions()
                ),
            ),
            req,
        )

    def tearDown(self):
        super(PackerFieldPermissionsRWTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._test_resource_packer

    def test_pack(self):
        new_data = TestData()
        expected_data = {"field2": 2, "field3": 3, "field4": 4}

        result = self._test_resource_packer.pack(new_data)
        self.assertDictEqual(result, expected_data)

    def test_unpack(self):
        new_data = {"field1": None, "field2": 2}

        result = self._test_resource_packer.unpack(new_data)
        self.assertDictEqual(result, new_data)


class JSONPackerIncludeNullTestCase(base.BaseTestCase):
    def setUp(self):
        req = mock.Mock()
        req.context.roles = ["owner"]

        super(JSONPackerIncludeNullTestCase, self).setUp()
        self._test_resource_packer = packers.JSONPackerIncludeNullFields(
            resources.ResourceByRAModel(
                FakeModel,
                fields_permissions=field_permissions.FieldsPermissionsByRole(
                    default=field_permissions.UniversalPermissions()
                ),
            ),
            req,
        )

    def tearDown(self):
        super(JSONPackerIncludeNullTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._test_resource_packer

    def test_pack(self):
        new_data = TestData()
        expected_data = {
            "field1": None,
            "field2": 2,
            "field3": 3,
            "field4": 4,
            "uuid": None,
        }

        result = orjson.loads(self._test_resource_packer.pack(new_data))
        self.assertDictEqual(result, expected_data)

    def test_unpack(self):
        new_data = {"field1": None, "field2": 2}

        result = self._test_resource_packer.unpack(
            orjson.dumps(new_data, option=orjson.OPT_NON_STR_KEYS)
        )
        self.assertDictEqual(result, new_data)


class MultipartPackerTestCase(base.BaseTestCase):
    _raw_http_request = (
        "POST /v1/docs/5fc2e03d-8b22-4baf-b16d-772c373b98e1/files/ "
        "HTTP/1.1\r\n"
        "Accept: */*\r\n"
        "Content-Length: 200\r\n"
        "Content-Type: multipart/form-data; "
        "boundary=------------------------hSlQJvPejd4JFNPeCJtXm0\r\n"
        "Host: 127.0.0.1:8080\r\n"
        "User-Agent: curl/8.12.1-DEV\r\n"
        "\r\n"
        "--------------------------hSlQJvPejd4JFNPeCJtXm0\r\n"
        'Content-Disposition: form-data; name="data"; filename="test.md"\r\n'
        "Content-Type: */*\r\n"
        "\r\n"
        "test_body\n"
        "\r\n"
        "--------------------------hSlQJvPejd4JFNPeCJtXm0--\r\n"
    )

    def setUp(self):
        super().setUp()
        self._req = webob.Request.from_text(self._raw_http_request)
        self._packer = packers.MultipartPacker(
            resources.ResourceByRAModel(FakeModel),
            self._req,
        )

    def tearDown(self):
        super().tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._packer

    def test_pack(self):
        new_data = b"test"

        result = self._packer.pack(new_data)
        assert result == new_data

    def test_unpack(self):
        result = self._packer.unpack(None)
        assert packers.MultipartPacker._multipart_key in result
        assert len(result[packers.MultipartPacker._parts_key]) == 1
        assert (
            next(iter(result[packers.MultipartPacker._parts_key]["data"].file))
            == b"test_body\n"
        )



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class RestAlchemyException(Exception):
    """Base REST Alchemy Exception.

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """

    message = "An unknown exception occurred."
    code = 500

    def __init__(self, **kwargs):
        self.msg = self.message % kwargs
        super(RestAlchemyException, self).__init__(self.msg)

    def __repr__(self):
        return "Code: %s, Message: %s" % (self.code, self.msg)

    def get_code(self):
        """
        This method is used in exception2dict function
        """
        return self.code


class PropertyNotFoundError(RestAlchemyException):
    message = "'%(class_name)s' object has no property '%(property_name)s'."
    code = 400


class NotFoundError(RestAlchemyException):
    message = "Nothing is found on path '%(path)s'."
    code = 404


class NotImplementedError(RestAlchemyException):
    base_message = "Not implemented."
    message = "Not implemented ('%(msg)s')."
    code = 501

    def __init__(self, **kwargs):
        if "msg" not in kwargs:
            self.message = self.base_message
        super(NotImplementedError, self).__init__(**kwargs)


class UnsupportedHttpMethod(RestAlchemyException):
    message = "HTTP method '%(method)s' is not supported."
    code = 405


class UnsupportedMethod(NotFoundError):
    message = "Method '%(method)s' is not supported for %(object_name)s."


class LocatorNotFound(NotFoundError):
    message = "Locator is not found for URI %(uri)s. Thus resource could not be found."


class UnknownResourceLocation(NotFoundError):
    message = (
        "Can not construct resource location for resource %(resource)r "
        "because the resource can't be got using REST API."
    )


class CanNotFindResourceByModel(NotFoundError):
    message = "Can not find a resource by model (%(model)r)."


class IncorrectRouteAttributeClass(NotFoundError):
    message = "Route %(route)s is of unacceptable class."


class IncorrectRouteAttribute(NotFoundError):
    message = "Route %(route)s doesn't have method %(attr)s."


class IncorrectActionCall(NotFoundError):
    message = "Action %(action)s is incorrectly called with HTTP method %(method)s."


class ResourceNotFoundError(NotFoundError):
    message = "Resource '%(resource)s' is not found on path: %(path)s."


class CollectionNotFoundError(NotFoundError):
    message = "Collection '%(collection)s' is not found on path: %(path)s."


class PropertyException(RestAlchemyException, ValueError):
    code = 400

    def __init__(self, name=None, model=None):
        self.name = name or "Unknown"
        self.model = model or "Unknown"
        super(PropertyException, self).__init__(name=self.name, model=self.model)


class PropertyRequired(PropertyException):
    message = (
        "Value for property '%(name)s' for model %(model)s "
        "is required! Property should not be None value."
    )


class ReadOnlyProperty(PropertyException):
    message = "Property '%(name)s' of model %(model)s is read only."


class TypeError(RestAlchemyException, TypeError):
    message = "Invalid type value '%(value)s' for '%(property_type)s'."
    code = 400

    def __init__(self, value, property_type):
        self._value = value
        self._property_type = property_type

        # Consider the nested type in output message
        if hasattr(property_type, "nested_type"):
            outer_type = type(property_type).__name__
            inner_type = type(property_type.nested_type).__name__
            property_type = f"{outer_type}({inner_type})"
        else:
            property_type = type(property_type).__name__

        super(TypeError, self).__init__(value=value, property_type=property_type)

    def get_value(self):
        return self._value

    def get_property_type(self):
        return self._property_type


class ModelTypeError(TypeError):
    message = (
        "Invalid type value '%(value)s'(%(value_type)s) for "
        "'%(model_name)s.%(property_name)s'(%(property_type)s)."
    )

    def __init__(self, value, property_name, property_type, model):
        super(TypeError, self).__init__(
            value=value,
            value_type=type(value),
            property_type=type(property_type).__name__,
            model_name=type(model).__name__,
            property_name=property_name,
        )


class RelationshipModelError(RestAlchemyException):
    message = (
        "Invalid model %(model)s for relationship. Must be inherited "
        "from dm.core.models.Model."
    )
    code = 400


class NotFoundOperationalStorageError(RestAlchemyException):
    message = "Can't get data for %(name)s key."


class ParseError(RestAlchemyException):
    message = "Can't parse value: %(value)s."
    code = 400

    def __init__(self, **kwargs):
        self.value = kwargs.get("value", "unknown")
        super().__init__(**kwargs)


class ParseBodyError(RestAlchemyException):
    message = "Can't parse body with specified content-type."
    code = 400


class FieldPermissionError(RestAlchemyException):
    message = "Permission denied for field %(field)s."
    code = 403


class ValidationErrorException(RestAlchemyException):
    """Base Validation Exception."""

    message = "Validation error occurred."
    code = 400


class ValidationPropertyPrivateError(ValidationErrorException):
    message = "Property %(property)s is private"


class ValidationPropertyIncompatibleError(ValidationErrorException):
    message = "%(val)s is not compatible with model %(model)s"


class ValidationFilterIncompatibleError(ValidationErrorException):
    message = "Filter %(val)s is not supported"


class ValidationSortInvalidDirValueError(ValidationErrorException):
    message = "Direction %(dir)s is not valid for sorting."


class ValidationSortIncompatibleDirCountError(ValidationErrorException):
    message = "Number of sort keys and directions is not equal."


class ValidationSortNumberError(ValidationErrorException):
    message = "Only one field can be sorted with such request."


class NotEqualUuidException(RestAlchemyException):
    message = (
        "Uuid (%(uuid)s) in body is not equal to parsed id (%(parsed_id)s) from url."
    )
    code = 400


class NotExtended(RestAlchemyException):
    message = "Application does not initialized with openapi engine."
    code = 400



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import orm
from restalchemy.tests.functional import base


class BatchDeleteModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "batch_delete_two_pk"
    foo_field1 = properties.property(types.Integer(), required=True, id_property=True)
    foo_field2 = properties.property(types.String(), default="foo_str")

    @property
    def super_id(self):
        return "%s-%s" % (self.uuid, self.foo_field1)

    @classmethod
    def get_id_property(cls):
        return {
            "super_id": (
                cls.properties["uuid"],
                cls.properties["foo_field1"],
            )
        }


class WithDbMigrationsDeleteTwoPkTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    def test_correct_batch_delete(self):
        my_models = BatchDeleteModel.objects.get_all()
        target = [my_models.pop(0), my_models.pop(2)]

        with self.engine.session_manager() as session:
            session.batch_delete(my_models)
        result = BatchDeleteModel.objects.get_all()

        self.assertEqual(target, result)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from restalchemy.api import middlewares
from restalchemy.tests.functional.restapi.ra_based.microservice import contexts


class ContextMiddleware(middlewares.ContextMiddleware):
    def process_request(self, req):
        ctx = contexts.Context()
        req.context = ctx
        result = req.get_response(self.application)
        ctx.release()
        return result



# Copyright 2021 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types as ra_types
from restalchemy.storage.sql import orm
from restalchemy.tests.functional import base

FAKE_UUID_1 = uuid.UUID("63cf2e1a-4f1f-11ec-8f05-1bfa6ad82a13")
FAKE_UUID_2 = uuid.UUID("78e4a492-4f1f-11ec-abd3-c362c434cef3")


class BinaryField(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "binary_data"

    data = properties.property(ra_types.String(), required=True)


class Bug29808262TestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "0001-test-bug-29808262-743d38"
    __FIRST_MIGRATION__ = "0001-test-bug-29808262-743d38.py"

    def test_blob_like_string(self):
        target = BinaryField(
            uuid=FAKE_UUID_1,
            data="aaa" * 1024,
        )
        target.insert()

        result = BinaryField.objects.get_one()

        self.assertEqual(result.data, target.data)

    def test_blob_like_float_bug29808262(self):
        target = BinaryField(
            uuid=FAKE_UUID_1,
            data="2" * 1024,
        )
        target.insert()

        # if bug is reproduced you will show like:
        # TypeError: Invalid type value 'inf' for 'String'
        result = BinaryField.objects.get_one()

        self.assertEqual(result.data, target.data)



# Copyright 2021 Eugene Frolov.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
import uuid

from parameterized import parameterized

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.tests import fixtures
from restalchemy.storage.sql.dialect.query_builder import q
from restalchemy.storage.sql import orm
from restalchemy.tests.utils import make_test_name

FAKE_UUID0 = uuid.UUID("00000000-0000-0000-0000-000000000000")
FAKE_UUID1 = uuid.UUID("00000000-0000-0000-0000-000000000001")
FAKE_UUID2 = uuid.UUID("00000000-0000-0000-0000-000000000002")


class SimpleModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "simple_table"

    field_str = properties.property(types.String(), default="FAKE_STR")
    field_int = properties.property(types.Integer(), default=1)
    field_bool = properties.property(types.Boolean(), default=True)


class ModelWithL1Relationships(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "model_with_l1_relationships"

    ref_l0_1 = relationships.relationship(SimpleModel, prefetch=True)
    ref_l0_2 = relationships.relationship(SimpleModel, prefetch=False)


class ModelWithL2Relationships(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "model_with_l2_relationships"

    ref_l1_1 = relationships.relationship(ModelWithL1Relationships, prefetch=True)
    ref_l1_2 = relationships.relationship(SimpleModel, prefetch=True)
    ref_l1_3 = relationships.relationship(ModelWithL1Relationships, prefetch=False)


# NOTE(efrolov): Sort model properties for correct ordering in asserts
for model in [SimpleModel, ModelWithL1Relationships, ModelWithL2Relationships]:
    model.properties.sort_properties()


class MySQLPrefetchQueryBuilderTestCase(unittest.TestCase):
    def setUp(self):
        super(MySQLPrefetchQueryBuilderTestCase, self).setUp()
        self.Q = q.Q
        model1 = SimpleModel(uuid=FAKE_UUID1)
        model2 = SimpleModel(uuid=FAKE_UUID2)
        self.flt = filters.AND(
            {"uuid": filters.EQ(FAKE_UUID0)},
            {"ref_l0_1": filters.LE(model1)},
            {"ref_l0_2": filters.NE(model2)},
        )

    def tearDown(self):
        super(MySQLPrefetchQueryBuilderTestCase, self).tearDown()
        del self.Q

    def test_l1_select(self):
        query = self.Q.select(
            model=ModelWithL1Relationships,
            session=fixtures.SessionFixture(),
        )

        result = query.compile()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`)",
            result,
        )

    def test_l1_select_with_filters(self):
        query = self.Q.select(
            model=ModelWithL1Relationships,
            session=fixtures.SessionFixture(),
        ).where(self.flt)

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "WHERE"
            " (`t1`.`uuid` = %s AND"
            " `t1`.`ref_l0_1` <= %s AND"
            " `t1`.`ref_l0_2` <> %s)",
            result_expression,
        )
        self.assertEqual(
            [
                "00000000-0000-0000-0000-000000000000",
                "00000000-0000-0000-0000-000000000001",
                "00000000-0000-0000-0000-000000000002",
            ],
            result_values,
        )

    def test_select_with_limit(self):
        query = self.Q.select(
            model=ModelWithL1Relationships,
            session=fixtures.SessionFixture(),
        ).limit(100500)

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "LIMIT 100500",
            result_expression,
        )
        self.assertEqual([], result_values)

    def test_l1_select_lock_with_filters(self):
        query = (
            self.Q.select(
                model=ModelWithL1Relationships,
                session=fixtures.SessionFixture(),
            )
            .where(self.flt)
            .for_()
        )

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            "WHERE"
            " (`t1`.`uuid` = %s AND"
            " `t1`.`ref_l0_1` <= %s AND"
            " `t1`.`ref_l0_2` <> %s) "
            "FOR UPDATE",
            result_expression,
        )
        self.assertEqual(
            [
                "00000000-0000-0000-0000-000000000000",
                "00000000-0000-0000-0000-000000000001",
                "00000000-0000-0000-0000-000000000002",
            ],
            result_values,
        )

    @parameterized.expand(
        [
            (None, "`t1`.`uuid` ASC"),
            ("ASC", "`t1`.`uuid` ASC"),
            ("DESC", "`t1`.`uuid` DESC"),
            (
                "ASC NULLS FIRST",
                "CASE WHEN `t1`.`uuid` IS NULL THEN 0 ELSE 1 END ASC, `t1`.`uuid` ASC",
            ),
            (
                "ASC NULLS LAST",
                "CASE WHEN `t1`.`uuid` IS NULL THEN 1 ELSE 0 END ASC, `t1`.`uuid` ASC",
            ),
            (
                "DESC NULLS FIRST",
                "CASE WHEN `t1`.`uuid` IS NULL THEN 0 ELSE 1 END ASC, `t1`.`uuid` DESC",
            ),
            (
                "DESC NULLS LAST",
                "CASE WHEN `t1`.`uuid` IS NULL THEN 1 ELSE 0 END ASC, `t1`.`uuid` DESC",
            ),
        ],
        name_func=make_test_name,
    )
    def test_l1_select_order_by(self, sort_dir, correct_clause):
        query = self.Q.select(
            model=ModelWithL1Relationships,
            session=fixtures.SessionFixture(),
        ).order_by("uuid", sort_dir)

        result_expression = query.compile()
        result_values = query.values()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l0_2` AS `t1_ref_l0_2`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`field_bool` AS `t2_field_bool`,"
            " `t2`.`field_int` AS `t2_field_int`,"
            " `t2`.`field_str` AS `t2_field_str`,"
            " `t2`.`uuid` AS `t2_uuid`"
            " FROM"
            " `model_with_l1_relationships` AS `t1` "
            "LEFT JOIN"
            " `simple_table` AS `t2` "
            "ON"
            " (`t1`.`ref_l0_1` = `t2`.`uuid`) "
            f"ORDER BY {correct_clause}",
            result_expression,
        )
        self.assertEqual([], result_values)

    def test_l2_select(self):
        result = self.Q.select(
            model=ModelWithL2Relationships,
            session=fixtures.SessionFixture(),
        ).compile()

        self.assertEqual(
            "SELECT"
            " `t1`.`ref_l1_3` AS `t1_ref_l1_3`,"
            " `t1`.`uuid` AS `t1_uuid`,"
            " `t2`.`ref_l0_2` AS `t2_ref_l0_2`,"
            " `t2`.`uuid` AS `t2_uuid`,"
            " `t3`.`field_bool` AS `t3_field_bool`,"
            " `t3`.`field_int` AS `t3_field_int`,"
            " `t3`.`field_str` AS `t3_field_str`,"
            " `t3`.`uuid` AS `t3_uuid`,"
            " `t4`.`field_bool` AS `t4_field_bool`,"
            " `t4`.`field_int` AS `t4_field_int`,"
            " `t4`.`field_str` AS `t4_field_str`,"
            " `t4`.`uuid` AS `t4_uuid` "
            "FROM"
            " `model_with_l2_relationships` AS `t1` "
            "LEFT JOIN"
            " `model_with_l1_relationships` AS `t2` "
            "ON"
            " (`t1`.`ref_l1_1` = `t2`.`uuid`) "
            "LEFT JOIN"
            " `simple_table` AS `t3` "
            "ON"
            " (`t2`.`ref_l0_1` = `t3`.`uuid`) "
            "LEFT JOIN"
            " `simple_table` AS `t4` "
            "ON"
            " (`t1`.`ref_l1_2` = `t4`.`uuid`)",
            result,
        )


class MySQLResultParserTestCase(unittest.TestCase):
    def test_l1_prefetch_result_parser(self):
        row_from_db = {
            "t1_ref_l0_2": "FakeRsUUID",
            "t1_uuid": "FakeUUID0",
            "t2_field_bool": "FakeBool",
            "t2_field_int": "FakeInt",
            "t2_field_str": "FakeStr",
            "t2_uuid": "FakeUUID1",
        }
        select_clause = q.Q.select(
            ModelWithL1Relationships,
            fixtures.SessionFixture(),
        )

        result = select_clause.parse_row(row_from_db)

        self.assertEqual(
            {
                "uuid": "FakeUUID0",
                "ref_l0_1": {
                    "field_bool": "FakeBool",
                    "field_int": "FakeInt",
                    "field_str": "FakeStr",
                    "uuid": "FakeUUID1",
                },
                "ref_l0_2": "FakeRsUUID",
            },
            result,
        )



# Copyright 2020 Dima Burmistrov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from restalchemy.dm import filters
from restalchemy.tests.unit import base


class FilterEqualityTestCase(base.BaseTestCase):
    def test_filters_equal(self):
        f1 = filters.EQ(4)
        f2 = filters.EQ(4)

        self.assertEqual(f1, f2)

    def test_filters_str_repr(self):
        random_uuid = uuid.uuid4()
        f = filters.EQ(random_uuid)
        self.assertEqual(str(f), str(random_uuid))

    def test_filters_not_equal_type(self):
        f1 = filters.GT(4)
        f2 = filters.EQ(4)

        self.assertNotEqual(f1, f2)

    def test_filters_not_equal_value(self):
        f1 = filters.GT(4)
        f2 = filters.GT(-80)

        self.assertNotEqual(f1, f2)

    def test_expr_filters_equal(self):
        f1 = filters.AND([filters.EQ(4), filters.NE(10)])
        f2 = filters.AND([filters.EQ(4), filters.NE(10)])

        self.assertEqual(f1, f2)

    def test_expr_filters_not_equal_type(self):
        f1 = filters.AND([filters.EQ(4), filters.NE(10)])
        f2 = filters.AND([filters.GT(4), filters.NE(10)])

        self.assertNotEqual(f1, f2)

    def test_expr_filters_not_equal_value(self):
        f1 = filters.AND([filters.EQ(2), filters.NE(10)])
        f2 = filters.AND([filters.EQ(4), filters.NE(10)])

        self.assertNotEqual(f1, f2)



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
# Copyright 2025 Genesis Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.common import exceptions
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.tests.unit import base

FAKE_VALUE = "FAKE_VALUE"
FAKE_VALUE2 = "FAKE_VALUE2"
FAKE_VALUE3 = "FAKE_VALUE3"


class PropertyTestCase(base.BaseTestCase):
    def setUp(self):
        super(PropertyTestCase, self).setUp()
        self.positive_fake_property_type = mock.MagicMock(
            **{
                "validate": mock.MagicMock(return_value=True),
                "spec": types.BaseType,
            }
        )
        self.negative_fake_property_type = mock.MagicMock(
            **{
                "validate": mock.MagicMock(return_value=False),
                "spec": types.BaseType,
            }
        )

    def _set_property_value(self, obj, value):
        obj.value = value

    def test_init_with_correct_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, value=FAKE_VALUE
        )

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_init_incorrect_value(self):
        self.assertRaises(
            exceptions.TypeError,
            properties.Property,
            self.negative_fake_property_type,
            value=FAKE_VALUE,
        )

    def test_init_default_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, default=FAKE_VALUE
        )

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_init_default_callable_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, default=lambda: FAKE_VALUE
        )

        self.assertEqual(property_obj.value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_init_default_value_if_property_read_only(self):
        property_obj = properties.Property(
            self.positive_fake_property_type,
            default=FAKE_VALUE,
            read_only=True,
        )

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_init_none_and_require(self):
        self.assertRaises(
            exceptions.PropertyRequired,
            properties.Property,
            self.negative_fake_property_type,
            required=True,
            value=None,
        )

    def test_init_value_is_none_default_is_not_none_and_prop_required(self):
        property_obj = properties.Property(
            self.positive_fake_property_type,
            value=None,
            required=True,
            default=FAKE_VALUE,
        )

        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_init_none_value(self):
        property_obj = properties.Property(self.negative_fake_property_type, value=None)

        self.assertEqual(property_obj._value, None)

    def test_init_value_with_read_only(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, read_only=True, value=FAKE_VALUE2
        )

        self.assertEqual(property_obj._value, FAKE_VALUE2)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE2)

    def test_set_correct_value(self):
        property_obj = properties.Property(self.positive_fake_property_type)

        self.assertIsNone(self._set_property_value(property_obj, FAKE_VALUE))
        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_set_incorect_value(self):
        property_obj = properties.Property(self.negative_fake_property_type)
        old_value = property_obj._value

        self.assertRaises(
            exceptions.TypeError,
            self._set_property_value,
            property_obj,
            FAKE_VALUE,
        )
        self.assertEqual(property_obj._value, old_value)
        self.negative_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_set_value_if_property_read_only(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, read_only=True
        )

        self.assertRaises(
            exceptions.ReadOnlyProperty,
            self._set_property_value,
            property_obj,
            FAKE_VALUE2,
        )
        self.assertEqual(property_obj._value, None)

    def test_set_value_if_property_read_only_and_value_is_same(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, read_only=True, value=FAKE_VALUE
        )

        self.assertIsNone(self._set_property_value(property_obj, FAKE_VALUE))
        self.assertEqual(property_obj._value, FAKE_VALUE)

    def test_set_value_if_property_read_only_and_value_is_different(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, read_only=True, value=FAKE_VALUE
        )

        self.assertRaises(
            exceptions.ReadOnlyProperty,
            self._set_property_value,
            property_obj,
            FAKE_VALUE2,
        )
        self.assertEqual(property_obj._value, FAKE_VALUE)

    def test_set_value_if_id_property_and_value_is_same(self):
        property_obj = properties.IDProperty(
            self.positive_fake_property_type, value=FAKE_VALUE
        )

        self.assertIsNone(self._set_property_value(property_obj, FAKE_VALUE))
        self.assertEqual(property_obj._value, FAKE_VALUE)

    def test_set_value_if_id_property_and_value_is_different(self):
        property_obj = properties.IDProperty(
            self.positive_fake_property_type, value=FAKE_VALUE
        )

        self.assertRaises(
            exceptions.ReadOnlyProperty,
            self._set_property_value,
            property_obj,
            FAKE_VALUE2,
        )
        self.assertEqual(property_obj._value, FAKE_VALUE)

    def test_set_force_correct_value(self):
        property_obj = properties.Property(self.positive_fake_property_type)

        self.assertIsNone(property_obj.set_value_force(FAKE_VALUE))
        self.assertEqual(property_obj._value, FAKE_VALUE)
        self.positive_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_set_force_incorect_value(self):
        property_obj = properties.Property(self.negative_fake_property_type)
        old_value = property_obj._value

        self.assertRaises(
            exceptions.TypeError, property_obj.set_value_force, FAKE_VALUE
        )
        self.assertEqual(property_obj._value, old_value)
        self.negative_fake_property_type.validate.assert_called_with(FAKE_VALUE)

    def test_set_force_none_value(self):
        property_obj = properties.Property(
            self.positive_fake_property_type, required=True, value=FAKE_VALUE
        )
        old_value = property_obj._value

        self.assertRaises(
            exceptions.PropertyRequired, property_obj.set_value_force, None
        )
        self.assertEqual(property_obj._value, old_value)


class PropertyCreatorTestCase(base.BaseTestCase):
    ARGS = [1, 2, 3]
    KWARGS = {"fake_key1": "fake_value1", "fake_key2": "fake_value2"}

    def setUp(self):
        self.property_mock = mock.Mock(return_value=FAKE_VALUE)
        self.prop_type_mock = mock.Mock()
        self.test_instance = properties.PropertyCreator(
            self.property_mock, self.prop_type_mock, self.ARGS, self.KWARGS
        )

    def test_call_object(self):
        self.assertEqual(self.test_instance(FAKE_VALUE2), FAKE_VALUE)
        self.property_mock.assert_called_once_with(
            value=FAKE_VALUE2,
            property_type=self.prop_type_mock,
            *self.ARGS,
            **self.KWARGS,
        )

    def test_get_property_class(self):
        self.assertEqual(self.test_instance.get_property_class(), self.property_mock)


class PropertyCollectionTestCase(base.BaseTestCase):
    def setUp(self):

        self.kwargs = {
            "fake1": mock.Mock(
                **{
                    "get_property_class.return_value": FAKE_VALUE,
                    "return_value": FAKE_VALUE2,
                }
            ),
            "fake2": mock.Mock(
                **{
                    "get_property_class.return_value": FAKE_VALUE2,
                    "return_value": FAKE_VALUE,
                }
            ),
        }
        self.test_instance = properties.PropertyCollection(**self.kwargs)

    def test_properties(self):
        self.assertEqual(self.test_instance.properties, self.kwargs)

    def test_change_properties_dict(self):
        def set_item(d, key, value):
            d[key] = value

        self.assertRaises(
            TypeError, set_item, self.test_instance.properties, "fake1", 2
        )

    def test_get_item_fake1(self):
        self.assertEqual(self.test_instance["fake1"], FAKE_VALUE)
        self.assertTrue(self.kwargs["fake1"].get_property_class.called)
        self.assertFalse(self.kwargs["fake2"].get_property_class.called)

    def test_get_item_fake2(self):
        self.assertEqual(self.test_instance["fake2"], FAKE_VALUE2)
        self.assertFalse(self.kwargs["fake1"].get_property_class.called)
        self.assertTrue(self.kwargs["fake2"].get_property_class.called)

    def test_instantiate_property_fake1(self):
        self.assertEqual(
            self.test_instance.instantiate_property(name="fake1", value=FAKE_VALUE3),
            FAKE_VALUE2,
        )
        self.kwargs["fake1"].assert_called_once_with(FAKE_VALUE3)
        self.assertFalse(self.kwargs["fake2"].called)

    def test_instantiate_property_fake2(self):
        self.assertEqual(
            self.test_instance.instantiate_property(name="fake2", value=FAKE_VALUE3),
            FAKE_VALUE,
        )
        self.kwargs["fake2"].assert_called_once_with(FAKE_VALUE3)
        self.assertFalse(self.kwargs["fake1"].called)

    def test_concatenate_two_collections(self):
        fake_property3 = mock.Mock()
        test_instance2 = properties.PropertyCollection(fake3=fake_property3)
        new_properties = self.kwargs
        new_properties["fake3"] = fake_property3

        res = self.test_instance + test_instance2
        self.assertIsInstance(res, properties.PropertyCollection)
        self.assertNotEqual(res, self.test_instance)
        self.assertNotEqual(res, test_instance2)
        self.assertEqual(res._properties, new_properties)

    def test_concatenate_collection_and_other_object(self):

        def concatenate(a, b):
            return a + b

        self.assertRaises(TypeError, concatenate, self.test_instance, object())


class PropertyManagerTestCase(base.BaseTestCase):
    def setUp(self):
        self.collection_mock = mock.Mock(
            **{
                "properties.items.return_value": [
                    ("fake1", "fake1"),
                    ("fake2", "fake2"),
                ],
                "instantiate_property.return_value": FAKE_VALUE,
            }
        )

    def test_init_manager(self):
        res = properties.PropertyManager(self.collection_mock)
        instantiate_property_calls = [
            mock.call("fake1", None),
            mock.call("fake2", None),
        ]

        self.assertIsInstance(res, properties.PropertyManager)
        self.collection_mock.instantiate_property.assert_has_calls(
            instantiate_property_calls, any_order=True
        )

    def test_init_manager_with_correct_kwargs(self):
        res = properties.PropertyManager(
            self.collection_mock, fake1=FAKE_VALUE2, fake2=FAKE_VALUE3
        )
        instantiate_property_calls = [
            mock.call("fake1", FAKE_VALUE2),
            mock.call("fake2", FAKE_VALUE3),
        ]

        self.assertIsInstance(res, properties.PropertyManager)
        self.collection_mock.instantiate_property.assert_has_calls(
            instantiate_property_calls, any_order=True
        )

    @base.unittest.skip("Checking of redundant data is turned off.")
    def test_init_manager_with_incorrect_kwargs(self):
        self.assertRaises(
            TypeError,
            properties.PropertyManager,
            self.collection_mock,
            fake3=FAKE_VALUE3,
        )

    def test_properties(self):
        property_manager = properties.PropertyManager(self.collection_mock)

        self.assertEqual(
            property_manager.properties,
            {"fake1": "FAKE_VALUE", "fake2": "FAKE_VALUE"},
        )

    def test_change_properties_dict(self):
        property_manager = properties.PropertyManager(self.collection_mock)

        def set_item(d, key, value):
            d[key] = value

        self.assertRaises(TypeError, set_item, property_manager.properties, "fake1", 2)


@mock.patch("restalchemy.dm.properties.PropertyCreator", return_value=FAKE_VALUE)
class PropertyFuncTestCase(base.BaseTestCase):
    ARGS = (1, 2, 3)
    KWARGS = {"fake_key1": "fake_value1", "fake_key2": "fake_value2"}

    def test_create_property(self, pc_mock):
        self.assertEqual(properties.property(*self.ARGS, **self.KWARGS), FAKE_VALUE)
        pc_mock.assert_called_once_with(
            prop_class=properties.Property,
            prop_type=1,
            args=self.ARGS[1:],
            kwargs=self.KWARGS,
        )

    def test_create_property_with_property_class(self, pc_mock):

        class LocalProperty(properties.AbstractProperty):
            @property
            def value(self):
                pass

            def set_value_force(self, value):
                pass

        self.assertEqual(
            properties.property(
                property_class=LocalProperty, *self.ARGS, **self.KWARGS
            ),
            FAKE_VALUE,
        )
        pc_mock.assert_called_once_with(
            prop_class=LocalProperty,
            prop_type=1,
            args=self.ARGS[1:],
            kwargs=self.KWARGS,
        )

    def test_create_property_with_incorect_property_class(self, pc_mock):
        self.assertRaises(
            ValueError,
            properties.property,
            property_class=object,
            *self.ARGS,
            **self.KWARGS,
        )



# Copyright (c) 2022 George Melikov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import logging

from restalchemy.api import middlewares


class Http500FakeResponse(object):
    def __init__(self, content_length=0, status_code=500):
        self.content_length = content_length
        self.status_code = status_code


class LoggingHttpMiddleware(middlewares.Middleware):
    """API HTTP logging middleware."""

    def __init__(self, application, logger_name=__name__):
        super(LoggingHttpMiddleware, self).__init__(application)
        self.logger = logging.getLogger(logger_name)

    @staticmethod
    def _make_message(start, req, res):
        referer = req.referer or "-"
        duration = datetime.datetime.now() - start
        td_ms = int(duration.total_seconds() * 1000)

        return '%s "%s %s" %s %s "%s" "%s" %s' % (
            req.client_addr,
            req.method,
            req.url,
            res.status_code,
            res.content_length,
            referer,
            req.user_agent,
            td_ms,
        )

    def process_request(self, req):
        start = datetime.datetime.now()

        try:
            res = req.get_response(self.application)
        except Exception:
            res = Http500FakeResponse()
            raise
        else:
            return res
        finally:
            msg = self._make_message(start, req, res)
            self.logger.info(msg)



# Copyright 2017 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import contextlib
import logging

from restalchemy.common import contexts
from restalchemy.storage.sql.dialect import mysql as mysql_dialect
from restalchemy.storage.sql.dialect import pgsql as pgsql_dialect

LOG = logging.getLogger(__name__)
DEFAULT_SAVEPOINT_NAME = "default_savepoint"


def escape(value):
    return "`%s`" % value


@contextlib.contextmanager
def savepoint(name: str = DEFAULT_SAVEPOINT_NAME):
    """Context manager that creates a savepoint and rolls back to it on error.

    The function can be used as a decorator or a context manager. For example:

    @savepoint()
    def my_function():
        pass

    with savepoint() as session:
        pass
    """
    if not name.isidentifier():
        raise ValueError(f"Invalid savepoint name: {name}")

    ctx = contexts.Context()
    engine = ctx._engine
    dialect_name = engine.dialect.name

    if dialect_name == pgsql_dialect.PgSQLDialect.DIALECT_NAME:
        expression_map = pgsql_dialect.SAVEPOINT_EXP_MAP
    elif dialect_name == mysql_dialect.MySQLDialect.DIALECT_NAME:
        expression_map = mysql_dialect.SAVEPOINT_EXP_MAP
    else:
        raise ValueError("Unsupported database dialect: %s" % dialect_name)

    savepoint_exp = expression_map["savepoint"].format(name=name)
    rollback_exp = expression_map["rollback"].format(name=name)
    release_exp = expression_map["release"].format(name=name)

    session = ctx.get_session()
    session.execute(savepoint_exp, tuple())

    try:
        yield session
    except Exception:
        LOG.error("Exception occurred, rolling back to savepoint")
        session.execute(rollback_exp, tuple())
        raise
    finally:
        session.execute(release_exp, tuple())



# Copyright 2022 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.api import constants


class Permissions(object):
    __slots__ = ()
    HIDDEN = 1
    RO = READONLY = 2
    RW = READWRITE = 3

    ALL_PERMISSIONS = (
        HIDDEN,
        RO,
        RW,
    )


class BasePermissions(object):
    def __init__(self, permission=Permissions.RW):
        self._permission = permission

    def meets_field_permission(self, model_field_name, req, current_permission):
        raise NotImplementedError()

    def is_readonly(self, model_field_name, req):
        return self.meets_field_permission(
            model_field_name=model_field_name,
            req=req,
            current_permission=Permissions.RO,
        )

    def is_hidden(self, model_field_name, req):
        return self.meets_field_permission(
            model_field_name=model_field_name,
            req=req,
            current_permission=Permissions.HIDDEN,
        )


class UniversalPermissions(BasePermissions):
    def __init__(self, permission=Permissions.RW):
        """General fields permissions container for resource

        This class describes a dict of fields with one permission
        for all fields.

        Example of usage:

        ```
        UniversalPermissions(
                permission=Permissions.RW
            )
        ```
        This code set to all fields READWRITE permissions.
        """
        super(UniversalPermissions, self).__init__(permission)

    def meets_field_permission(self, model_field_name, req, current_permission):
        return self._permission <= current_permission


class FieldsPermissions(BasePermissions):
    def __init__(self, fields, default=Permissions.RW):
        """Field permissions container for resource

        This class describes a dict of fields with permissions for
        various RestAlchemy API methods. All permissions are decrared
        in class Permissions. The methods supported by RA are declared
        in the `restalchemy.api.constants` module.
        For example, the code below shows how to set:
        > HIDDEN permissions to the `one_field` for the FILTER method
          and RW for all others RA methods, except FILTER;
        > RO permissions for the `two_field` for all RA methods;
        > HIDDEN permissions for the `three_field` for GET method and
          RO permissions for all others RA methods, except GET.

        ```
        FieldsPermissions(
            default=Permissions.RW,
            fields={
                'one_field':{
                    constants.FILTER: Permissions.HIDDEN},
                'two_field': {
                    constants.ALL: Permissions.RO},
                'three_field': {
                    constants.GET: Permissions.HIDDEN,
                    constants.ALL: Permissions.RO},
            }
        )
        ```

        Pay attention: if you wouldn't set permission for field and RA method
        by default it would be RW (READWRITE) permission.

        :param fields: dict of field model name and permissions
        :param default: default permission for non-described fields
        """
        for method_permission in fields.values():
            for method, permission in method_permission.items():
                assert (
                    method.upper() in constants.ALL_RA_METHODS
                    and permission in Permissions.ALL_PERMISSIONS
                )
        super(FieldsPermissions, self).__init__(permission=default)
        self.fields = fields

    def meets_field_permission(self, model_field_name, req, current_permission):

        method = req.api_context.get_active_method()
        field_permission = self.fields.get(model_field_name, {})

        # NOTE(g.melikov): By DEFAULT permission is Permissions.RW
        permission = (
            field_permission.get(method)
            or field_permission.get(constants.ALL)
            or self._permission
        )

        return permission <= current_permission


class FieldsPermissionsByRole(BasePermissions):
    def __init__(self, default, **kwargs):
        """Role based fields permissions

        This class describes a dict of roles with FieldsPermissions.
        Note: default will be used for all roles which are not specified.

        Example of usage:

        ```
        FieldsPermissionsByRole(
        some_role=FieldsPermissions(
            default=Permissions.RW,
            fields={
                'status':
                {
                    constants.ALL: Permissions.RO
                }
            }
        ),
        admin=FieldsPermissions(
            default=Permissions.RW,
            fields={
                'namespace':
                {
                    constants.ALL: Permissions.HIDDEN
                }
            }
        ),
        default=UniversalPermissions(
            permission=Permissions.RW
            )
        )
        ```

        :param default: default permission for non-described roles
        :param kwargs: dict of roles with FieldsPermissions.
        """
        for role, permissions in kwargs.items():
            if not isinstance(permissions, BasePermissions):
                raise NotImplementedError(
                    "Permissions for %s must be inherited BasePermissions class" % role
                )
        self.default = default
        self.role_fields = kwargs

        super(FieldsPermissionsByRole, self).__init__()

    @staticmethod
    def _get_roles(req):
        return (
            req.context.roles
            if hasattr(req, "context") and hasattr(req.context, "roles")
            else []
        )

    def meets_field_permission(self, model_field_name, req, current_permission):

        for current_role in self._get_roles(req):
            if current_role in self.role_fields:
                fields = self.role_fields[current_role]
                return fields.meets_field_permission(
                    model_field_name, req, current_permission
                )

        return self.default.meets_field_permission(
            model_field_name, req, current_permission
        )



#    Copyright 2019 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import unittest

from restalchemy.storage.sql import migrations
from restalchemy.tests.functional import db_utils

INIT_MIGRATION = "9e335f-test-batch-migration"


class BaseFunctionalTestCase(unittest.TestCase):
    """All other functional tests should inherit from it."""

    pass


class BaseDBEngineTestCase(db_utils.DBEngineMixin, BaseFunctionalTestCase):
    """Base recommended class to inherit from for all db-related tests"""

    @classmethod
    def setUpClass(cls):
        super(BaseDBEngineTestCase, cls).setUpClass()

        cls.init_engine()

    @classmethod
    def tearDownClass(cls):
        super(BaseDBEngineTestCase, cls).tearDownClass()

        cls.drop_all_tables()
        cls.destroy_engine()


class BaseWithDbMigrationsTestCase(BaseDBEngineTestCase):
    __LAST_MIGRATION__ = None
    __FIRST_MIGRATION__ = None

    def setUp(self):
        super(BaseWithDbMigrationsTestCase, self).setUp()

        # configure database structure, apply migrations
        self._migrations = self.get_migration_engine()
        self._migrations.rollback_migration(self.__FIRST_MIGRATION__)
        self._migrations.apply_migration(self.__LAST_MIGRATION__)

    def tearDown(self):
        super(BaseWithDbMigrationsTestCase, self).tearDown()

        # destroy database structure, rollback migrations
        self._migrations = self.get_migration_engine()
        self._migrations.rollback_migration(self.__FIRST_MIGRATION__)

    @staticmethod
    def get_migration_engine():
        migrations_path = os.path.dirname(__file__) + "/migrations/"
        migration_engine = migrations.MigrationEngine(migrations_path=migrations_path)
        return migration_engine



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import copy
import inspect

import builtins
from collections import abc as collections_abc

from restalchemy.common import exceptions as exc
from restalchemy.common import utils
from restalchemy.dm import types


class AbstractProperty(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def value(self):
        pass

    @abc.abstractmethod
    def set_value_force(self, value):
        pass

    @abc.abstractmethod
    def is_dirty(self):
        pass

    @classmethod
    def is_prefetch(cls):
        return False


class BaseProperty(AbstractProperty):
    pass


class Property(BaseProperty):
    def __init__(
        self,
        property_type,
        default=None,
        required=False,
        read_only=False,
        value=None,
        mutable=False,
        example=None,
    ):
        if not isinstance(property_type, types.BaseType):
            raise TypeError("Property type must be instance of %s" % types.BaseType)
        self._type = property_type
        self._required = bool(required)
        self._read_only = bool(read_only)
        if value is not None:
            self.set_value_force(value)
        elif callable(default):
            self.set_value_force(default())
        else:
            self.set_value_force(default)
        self.__first_value = copy.deepcopy(self.value) if mutable else self.value
        self._example = example

    def is_dirty(self):
        return not self.__first_value == self.value

    def _safe_value(self, value):
        if value is None or self._type.validate(value):
            if value is None and self.is_required():
                raise exc.PropertyRequired()
            return value
        else:
            raise exc.TypeError(value=value, property_type=self._type)

    def is_read_only(self):
        return self._read_only

    def is_required(self):
        return self._required

    @classmethod
    def is_id_property(cls):
        return False

    @builtins.property
    def old_value(self):
        return self.__first_value

    @builtins.property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self.is_read_only() or self.is_id_property():
            if value != self._value:
                raise exc.ReadOnlyProperty()
        self._value = self._safe_value(value)

    def set_value_force(self, value):
        self._value = self._safe_value(value)

    @builtins.property
    def property_type(self):
        return self._type

    def get_property_type(self):
        return self._type

    def example(self):
        return self._example


class IDProperty(Property):
    @classmethod
    def is_id_property(cls):
        return True


class PropertyCreator(object):
    def __init__(self, prop_class, prop_type, args, kwargs):
        self._property = prop_class
        self._property_type = prop_type
        self._args = args
        self._kwargs = kwargs
        self._prefetch = kwargs.pop("prefetch", False)

    def __call__(self, value):
        return self._property(
            value=value, property_type=self._property_type, *self._args, **self._kwargs
        )

    def get_property_class(self):
        return self._property

    def get_property_type(self):
        return self._property_type

    def get_kwargs(self):
        return self._kwargs

    def is_prefetch(self):
        return self._prefetch


class PropertyMapping(collections_abc.Mapping, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def properties(self):
        pass

    def __getitem__(self, name):
        return self.properties[name]

    def __iter__(self):
        return iter(self.properties)

    def __len__(self):
        return len(self.properties)


class PropertyCollection(PropertyMapping):
    def __init__(self, **kwargs):
        self._properties = kwargs
        super(PropertyCollection, self).__init__()

    def sort_properties(self):
        """Switch the model to sorted properties

        After call this method all requests to Model.properties will return
        sorted values. For example:

        ```
        class Model(Model):
            b = properties.property(type.B())
            a = properties.property(type.A())

        Model.properties.keys()  #-> ['b', 'a']
        Model.properties.sort_properties()
        Model.properties.keys()  #-> ['a', 'b']
        ```

        Most often, this functionality is needed for tests.
        """
        result = collections.OrderedDict()
        for key in sorted(self._properties):
            result[key] = self._properties[key]
        self._properties = result

    def __getitem__(self, name):
        return self.properties[name].get_property_class()

    @builtins.property
    def properties(self):
        return utils.ReadOnlyDictProxy(self._properties)

    def __add__(self, other):
        if isinstance(other, PropertyCollection):
            props = dict(self.properties)
            props.update(other.properties)
            return type(self)(**props)
        raise TypeError(
            "Cannot concatenate %s and %s objects"
            % (type(self).__name__, type(other).__name__)
        )

    def instantiate_property(self, name, value=None):
        return self._properties[name](value)

    def get_property_class(self):
        return type(self)


class PropertyManager(PropertyMapping):
    def __init__(self, property_collection, **kwargs):
        self._properties = {}
        for name, item in property_collection.properties.items():
            if isinstance(item, PropertyCollection):
                prop = PropertyManager(item, **kwargs.pop(name, {}))
            else:
                try:
                    prop = property_collection.instantiate_property(
                        name, kwargs.pop(name, None)
                    )
                except exc.PropertyRequired:
                    raise exc.PropertyRequired(name=name)
            self._properties[name] = prop

        # commented because kwargs can contain 'context' etc. Figure out
        #        if len(kwargs) > 0:
        #            raise TypeError("Unknown parameters: %s" % str(kwargs))
        super(PropertyManager, self).__init__()

    @builtins.property
    def properties(self):
        return utils.ReadOnlyDictProxy(self._properties)

    @builtins.property
    def value(self):
        result = {}
        for k, v in self.properties.items():
            result[k] = v.value
        return result

    @value.setter
    def value(self, values):
        for k, v in values.items():
            self._properties[k].value = v


def property(property_type, *args, **kwargs):
    id_property = kwargs.pop("id_property", False)
    property_class = kwargs.pop(
        "property_class", IDProperty if id_property else Property
    )
    if inspect.isclass(property_class) and issubclass(property_class, AbstractProperty):
        return PropertyCreator(
            prop_class=property_class,
            prop_type=property_type,
            args=args,
            kwargs=kwargs,
        )
    else:
        raise ValueError(
            "Value of property class argument (%s) must be"
            " inherited on AbstractProperty class"
            "" % str(property_class)
        )


def container(**kwargs):
    kwargs = copy.deepcopy(kwargs)
    for prop in kwargs.values():
        if not isinstance(prop, (PropertyCreator, PropertyCollection)):
            raise Exception("Only property, relationship and container are allowed.")
    return PropertyCollection(**kwargs)


def required_property(property_type, *args, **kwargs):
    kwargs["required"] = True
    return property(property_type, *args, **kwargs)


def readonly_property(property_type, *args, **kwargs):
    kwargs["read_only"] = True
    return required_property(property_type, *args, **kwargs)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import logging
import threading

from mysql.connector import errors
from psycopg import rows as pg_rows
from psycopg import errors as pg_errors

from restalchemy.storage import exceptions as exc
from restalchemy.storage.sql.dialect import mysql
from restalchemy.storage.sql.dialect import pgsql

LOG = logging.getLogger(__name__)


class SessionQueryCache(object):
    def __init__(self, session):
        super(SessionQueryCache, self).__init__()
        self._session = session
        self.__query_cache = {}

    def _get_hash(
        self, engine, table, filters, limit=None, order_by=None, locked=False
    ):
        query = engine.dialect.select(
            table=table,
            filters=filters,
            limit=limit,
            order_by=order_by,
            session=self._session,
            locked=locked,
        )
        values = query.get_values()
        statement = query.get_statement()
        return hash(tuple([statement] + values))

    def _get_hash_by_query(
        self,
        engine,
        table,
        where_conditions,
        where_values,
        limit=None,
        order_by=None,
        locked=False,
    ):
        query = engine.dialect.custom_select(
            table=table,
            where_conditions=where_conditions,
            where_values=where_values,
            limit=limit,
            order_by=order_by,
            session=self._session,
            locked=locked,
        )
        values = query.get_values()
        statement = query.get_statement()
        return hash(tuple([statement] + values))

    def get_all(
        self,
        engine,
        table,
        filters,
        fallback,
        limit=None,
        order_by=None,
        locked=False,
    ):
        query_hash = self._get_hash(engine, table, filters, limit, locked)
        if query_hash not in self.__query_cache:
            self.__query_cache[query_hash] = fallback(
                filters=filters,
                session=self._session,
                limit=limit,
                order_by=order_by,
                locked=locked,
            )
        return self.__query_cache[query_hash]

    def query(
        self,
        engine,
        table,
        where_conditions,
        where_values,
        fallback,
        limit=None,
        order_by=None,
        locked=False,
    ):
        query_hash = self._get_hash_by_query(
            engine, table, where_conditions, where_values, limit, locked
        )
        if query_hash not in self.__query_cache:
            self.__query_cache[query_hash] = fallback(
                where_conditions=where_conditions,
                where_values=where_values,
                session=self._session,
                limit=limit,
                order_by=order_by,
                locked=locked,
            )
        return self.__query_cache[query_hash]


class PgSQLSession(object):
    def __init__(self, engine):
        self._engine = engine
        self._conn = self._engine.get_connection()
        self._cursor = self._conn.cursor(row_factory=pg_rows.dict_row)
        self._log = LOG
        self.cache = SessionQueryCache(session=self)

    @property
    def engine(self):
        return self._engine

    @staticmethod
    def _check_models_same_type(target_model, models):
        model_type = type(target_model)
        if not min(map(lambda m: isinstance(m, model_type), models)):
            raise TypeError("All models in the list must be of the same type")

    def batch_insert(self, models):
        if models:
            # Check models type
            first_model = models[0]
            self._check_models_same_type(first_model, models)

            # process values
            values = []
            table = first_model.get_table()
            statement = pgsql.PgSQLInsert(
                table=table,
                data=first_model.get_storable_snapshot(),
                session=self,
            ).get_statement()
            for model in models:
                snapshot = model.get_storable_snapshot()
                insert = pgsql.PgSQLInsert(
                    table=table,
                    data=snapshot,
                    session=self,
                )
                values.append(insert.get_values())

            try:
                return self.execute_many(statement, values)
            except pg_errors.UniqueViolation as e:
                raise exc.ConflictRecords(
                    model=type(first_model).__name__,
                    msg=str(e),
                )

    def batch_delete(self, models):
        if models:
            # Check models type
            first_model = models[0]
            self._check_models_same_type(first_model, models)

            # process values
            table = first_model.get_table()
            values = []
            for model in models:
                pk_values = {}
                for name, prop in model.get_id_properties().items():
                    pk_values[name] = prop.property_type.to_simple_type(prop.value)
                values.append(pk_values)
            operation = pgsql.PgSQLBatchDelete(
                table=table,
                snapshot=values,
                session=self,
            )

            return self.execute(operation.get_statement(), operation.get_values())

    def execute(self, statement, values=None):
        try:
            self._log.debug(
                ("Execute statement %s with values %s within %s database"),
                statement,
                values,
                self._engine.db_name,
            )
            self._cursor.execute(statement, values)
            return self._cursor
        except errors.DatabaseError:
            raise

    def execute_many(self, statement, values):
        self._log.debug(
            ("Execute batch statement %s with values %s within %s database"),
            statement,
            values,
            self._engine.db_name,
        )
        self._cursor.executemany(statement, values)
        return self._cursor

    def rollback(self):
        self._conn.rollback()

    def commit(self):
        self._conn.commit()

    def close(self):
        self._engine.close_connection(self._conn)


class MySQLSession(object):
    def __init__(self, engine):
        self._engine = engine
        self._conn = self._engine.get_connection()
        self._cursor = self._conn.cursor(dictionary=True, buffered=True)
        self._log = LOG
        self.cache = SessionQueryCache(session=self)

    @property
    def engine(self):
        return self._engine

    @staticmethod
    def _check_models_same_type(target_model, models):
        model_type = type(target_model)
        if not min(map(lambda m: isinstance(m, model_type), models)):
            raise TypeError("All models in the list must be of the same type")

    def batch_insert(self, models):
        if models:
            # Check models type
            first_model = models[0]
            self._check_models_same_type(first_model, models)

            # process values
            values = []
            table = first_model.get_table()
            statement = mysql.MySQLInsert(
                table=table,
                data=first_model.get_storable_snapshot(),
                session=self,
            ).get_statement()
            for model in models:
                snapshot = model.get_storable_snapshot()
                insert = mysql.MySQLInsert(
                    table=table,
                    data=snapshot,
                    session=self,
                )
                values.append(insert.get_values())

            try:
                return self.execute_many(statement, values)
            except errors.IntegrityError as e:
                # Error codes from Maria DB documentation. See more on website
                # https://mariadb.com/kb/en/mariadb-error-codes/
                # Errno: 1062 and sqlstate: 23000 is "Duplicate entry" error.
                if e.errno == 1062 and e.sqlstate == "23000":
                    raise exc.ConflictRecords(
                        model=type(first_model).__name__,
                        msg=e.msg,
                    )
                else:
                    raise exc.UnknownStorageException(caused=e)

    def batch_delete(self, models):
        if models:
            # Check models type
            first_model = models[0]
            self._check_models_same_type(first_model, models)

            # process values
            table = first_model.get_table()
            values = []
            for model in models:
                pk_values = {}
                for name, prop in model.get_id_properties().items():
                    pk_values[name] = prop.property_type.to_simple_type(prop.value)
                values.append(pk_values)
            operation = mysql.MySQLBatchDelete(
                table=table,
                snapshot=values,
                session=self,
            )

            return self.execute(operation.get_statement(), operation.get_values())

    def execute(self, statement, values=None):
        try:
            self._log.debug(
                ("Execute statement %s with values %s within %s database"),
                statement,
                values,
                self._engine.db_name,
            )
            self._cursor.execute(statement, values)
            return self._cursor
        except errors.DatabaseError as e:
            if e.errno == 1213:
                raise exc.DeadLock(msg=e.msg)
            raise

    def execute_many(self, statement, values):
        self._log.debug(
            ("Execute batch statement %s with values %s within %s database"),
            statement,
            values,
            self._engine.db_name,
        )
        self._cursor.executemany(statement, values)
        return self._cursor

    def rollback(self):
        self._conn.rollback()

    def commit(self):
        self._conn.commit()

    def close(self):
        self._engine.close_connection(self._conn)


@contextlib.contextmanager
def session_manager(engine, session=None):
    if session is None:
        session = engine.get_session()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()
    else:
        yield session


class SessionConflict(Exception):
    pass


class SessionNotFound(Exception):
    pass


class SessionThreadStorage(object):
    _storage = threading.local()

    def __init__(self):
        super(SessionThreadStorage, self).__init__()

    def get_session(self):
        thread_session = getattr(self._storage, "session", None)
        if thread_session is None:
            raise SessionNotFound("A session is not exists for this thread")
        return thread_session

    def pop_session(self):
        try:
            return self.get_session()
        finally:
            self.remove_session()

    def remove_session(self):
        self._storage.session = None

    def store_session(self, session):
        try:
            thread_session = self.get_session()
            raise SessionConflict(
                "Another session %r is already stored!", thread_session
            )
        except SessionNotFound:
            self._storage.session = session
            return self._storage.session



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
# Copyright 2014 Mirantis Inc
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import abc as collections_abc

from restalchemy.common import exceptions as ra_exc


class ReadOnlyDictProxy(collections_abc.Mapping):
    """A hashable, immutable dict proxy.

    Return a proxy object for a mapping which enforces read-only behavior.
    This is normally used to create a proxy to prevent modification of the
    dictionary for non-dynamic class types.
    """

    def __init__(self, d):
        self._d = d
        self._hash = None

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        if self._hash is None:
            self._hash = 0
            for key, value in self.items():
                self._hash ^= hash(key)
                self._hash ^= hash(value)

        return self._hash

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._d)


class classproperty(property):
    """Property for classes

    Standard property cannot be used on classes only on instances.
    This one was created to overcome this problem. The only limitation
    is that setter is not working for classes, only getter.

    This property could be used as:
        class A(object):
            @classproperty
            def test(cls):
                ...some cls processing...
                return result

        test = A.test  # method 'test' will be called with A class argument
    """

    def __get__(self, obj, cls):
        # Calls property function
        return self.fget(cls)


def lastslash(url):
    if url and url[-1] != "/":
        return url + "/"
    else:
        return url


def raise_parse_error_on_fail(func):
    """Decorator for handling model properties/relationships parsing errors

    Catches ValueError and TypeError exceptions during field validation and
    converts them into ParseError with HTTP 400 BadRequest status. Other
    exceptions are propagated to the error middleware for further processing.

    Any validation errors from controllers or properties related to validating
    the correctness of a value must be subclasses of ValueError or TypeError.
    """

    def wrapper(obj, name, value, *args, **kwargs):
        try:
            return func(obj, name, value, *args, **kwargs)
        except (ValueError, TypeError):
            raise ra_exc.ParseError(value="%s=%s" % (name, value))
        except ra_exc.ParseError as e:
            raise ra_exc.ParseError(value="%s=%s" % (name, e.value))

    return wrapper


def find_first(seq, predicate):
    """Return first item in seq for which predicate(item) is True"""
    return next((x for x in seq if predicate(x)), None)



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import logging

import webob

from restalchemy.api import constants
from restalchemy.api import packers
from restalchemy.api import resources
from restalchemy.common import exceptions as exc
from restalchemy.common import utils
from restalchemy.dm import filters as dm_filters
from restalchemy.openapi import constants as oa_c
from restalchemy.openapi import utils as oa_utils
from restalchemy.storage.sql import constants as sql_c

LOG = logging.getLogger(__name__)


class Controller(object):
    __resource__ = None  # type: resources.ResourceByRAModel

    # Not for common cases (Example: for JSONPackerIncludeNullFields)
    __packer__ = None  # type: packers.BaseResourcePacker

    # You can also generate location header for GET and UPDATE methods,
    # just expand the list with the following constants:
    #  * constants.GET
    #  * constants.UPDATE
    __generate_location_for__ = {
        constants.CREATE,
    }

    # List of fields that can be used for user sorting. If '__all__', any field is allowed.
    __sortable_fields__ = "__all__"

    # Default sorting, leave empty for no default sorting.
    __default_sort__ = {}

    def __init__(self, request):
        super(Controller, self).__init__()
        self._req = request

    def __repr__(self):
        return self.__class__.__name__

    @property
    def request(self):
        return self._req

    def get_packer(self, content_type, resource_type=None):
        packer = self.__packer__ or packers.get_packer(content_type)
        rt = resource_type or self.get_resource()
        return packer(rt, request=self._req)

    def _create_response(self, body, status, headers, charset="utf-8"):
        if body is not None:
            headers["Content-Type"] = packers.get_content_type(headers)
            packer = self.get_packer(headers["Content-Type"])
            body = packer.pack(body)

        return webob.Response(
            body=body,
            status=status,
            content_type=headers.get("Content-Type", None),
            headerlist=[(k, v) for k, v in headers.items()],
            charset=charset,
        )

    def process_result(self, result, status_code=200, headers=None, add_location=False):
        headers = headers or {}

        def correct(body, c=status_code, h=None, h_location=add_location, *args):
            h = h or {}
            if h_location:
                try:
                    headers["Location"] = resources.ResourceMap.get_location(body)
                except (
                    exc.UnknownResourceLocation,
                    exc.CanNotFindResourceByModel,
                ) as e:
                    LOG.warning("Can't construct location header by reason: %r", e)
            headers.update(h)
            return body, c, headers

        if isinstance(result, tuple):
            return self._create_response(*correct(*result))
        else:
            return self._create_response(*correct(result))

    def _make_kwargs(self, parent_resource, **kwargs):
        if parent_resource:
            kwargs["parent_resource"] = parent_resource
        return kwargs

    @utils.raise_parse_error_on_fail
    def _parse_field_value(self, name, value, resource_field):
        return resource_field.parse_value_from_unicode(self._req, value)

    def _prepare_filter(self, param_name, value):
        if self.model is None:
            raise exc.ValidationFilterIncompatibleError(val=param_name)
        try:
            resource_field = self.__resource__.get_field(
                self.__resource__.get_model_field_name(param_name)
            )
        except ValueError:
            raise exc.ValidationFilterIncompatibleError(val=param_name)

        value = self._parse_field_value(param_name, value, resource_field)

        return resource_field.name, value

    def _prepare_sorts(self, params):
        keys = params.getall("sort_key")
        if not keys:
            return self.__default_sort__
        if keys and self.__sortable_fields__ != "__all__":
            keys = [key for key in keys if key in self.__sortable_fields__]

        dirs = params.getall("sort_dir")
        for dir in dirs:
            if dir.upper() not in sql_c.SQL_SORT_SET:
                raise exc.ValidationSortInvalidDirValueError(dir=dir)
        if dirs:
            if len(keys) != len(dirs):
                raise exc.ValidationSortIncompatibleDirCountError()
            return dict(itertools.zip_longest(keys, dirs, fillvalue="asc"))

        return {key: "asc" for key in keys}

    def _prepare_filters(self, params):
        result = {}
        process = self.__resource__ and self.__resource__.is_process_filters()
        for param, value in params.items():
            filter_name, filter_value = (
                self._prepare_filter(param, value) if process else (param, value)
            )
            if filter_name not in result:
                result[filter_name] = dm_filters.EQ(filter_value)
            else:
                values = (
                    [result[filter_name].value]
                    if not isinstance(result[filter_name], dm_filters.In)
                    else result[filter_name].value
                )
                values.append(filter_value)
                result[filter_name] = dm_filters.In(values)

        return result

    def do_collection(self, parent_resource=None):
        method = self._req.method

        api_context = self._req.api_context
        if method == "GET":
            api_context.set_active_method(constants.FILTER)
            order_by = self._prepare_sorts(
                params=self._req.api_context.params,
            )
            filters = self._prepare_filters(
                params=self._req.api_context.params_filters,
            )
            kwargs = self._make_kwargs(parent_resource, filters=filters)
            return self.process_result(result=self.filter(**kwargs, order_by=order_by))
        elif method == "POST":
            api_context.set_active_method(constants.CREATE)
            content_type = packers.get_content_type(self._req.headers)
            packer = self.get_packer(content_type)
            kwargs = self._make_kwargs(
                parent_resource, **packer.unpack(value=self._req.body)
            )

            return self.process_result(
                result=self.create(**kwargs),
                status_code=201,
                add_location=constants.CREATE in self.__generate_location_for__,
            )
        else:
            raise exc.UnsupportedHttpMethod(method=method)

    def get_resource_by_uuid(self, uuid, parent_resource=None):
        kwargs = self._make_kwargs(parent_resource)

        parsed_id = (
            self._parse_resource_uuid("uuid", uuid, self.get_resource().get_id_type())
            if self.__resource__
            else uuid
        )

        result = self.get(uuid=parsed_id, **kwargs)
        if isinstance(result, tuple):
            return result[0]
        return result

    @utils.raise_parse_error_on_fail
    def _parse_resource_uuid(self, name, value, id_type):
        return id_type.from_unicode(value)

    def do_resource(self, uuid, parent_resource=None):
        method = self._req.method
        kwargs = self._make_kwargs(parent_resource)

        parsed_id = (
            self._parse_resource_uuid("uuid", uuid, self.get_resource().get_id_type())
            if self.__resource__
            else uuid
        )

        api_context = self._req.api_context

        if method == "GET":
            api_context.set_active_method(constants.GET)
            return self.process_result(
                result=self.get(uuid=parsed_id, **kwargs),
                add_location=constants.GET in self.__generate_location_for__,
            )
        elif method == "PUT":
            api_context.set_active_method(constants.UPDATE)
            content_type = packers.get_content_type(self._req.headers)
            packer = self.get_packer(content_type)
            kwargs.update(packer.unpack(value=self._req.body))
            # TODO(v.burygin) We need carefully change params of method update
            #  in new major version for updating uuid in model
            if "uuid" in kwargs:
                if kwargs["uuid"] != parsed_id:
                    raise exc.NotEqualUuidException(
                        uuid=kwargs["uuid"], parsed_id=parsed_id
                    )
                kwargs.pop("uuid", None)
            return self.process_result(
                result=self.update(uuid=parsed_id, **kwargs),
                add_location=constants.UPDATE in self.__generate_location_for__,
            )
        elif method == "DELETE":
            api_context.set_active_method(constants.DELETE)
            result = self.delete(uuid=parsed_id, **kwargs)
            return self.process_result(
                result=result,
                status_code=200 if result else 204,
            )
        else:
            raise exc.UnsupportedHttpMethod(method=method)

    @classmethod
    def get_resource(cls):
        return cls.__resource__

    @property
    def model(self):
        return self.get_resource().get_model()

    def create(self, **kwargs):
        raise exc.NotImplementedError(
            msg="method create in %s" % self.__class__.__name__
        )

    def get(self, uuid):
        raise exc.NotImplementedError(msg="method get in %s" % self.__class__.__name__)

    def filter(self, filters, order_by=None):
        raise exc.NotImplementedError(
            msg="method filter in %s" % self.__class__.__name__
        )

    def delete(self, uuid):
        raise exc.NotImplementedError(
            msg="method delete in %s" % self.__class__.__name__
        )

    def update(self, uuid, **kwargs):
        raise exc.NotImplementedError(
            msg="method update in %s" % self.__class__.__name__
        )

    def get_context(self):
        try:
            return self._req.context
        except AttributeError:
            return None


class BaseResourceController(Controller):
    def create(self, **kwargs):
        dm = self.model(**kwargs)
        dm.insert()
        return dm

    def get(self, uuid, **kwargs):
        # TODO(d.burmistrov): replace this hack with normal argument passing
        kwargs[self.model.get_id_property_name()] = dm_filters.EQ(uuid)
        return self.model.objects.get_one(filters=kwargs)

    def _split_filters(self, filters):
        if hasattr(self.model, "get_custom_properties"):
            custom_filters = {}
            storage_filters = {}
            custom_properties = dict(self.model.get_custom_properties())
            for name, value in filters.items():
                if name in custom_properties:
                    custom_filters[name] = value
                    continue
                storage_filters[name] = value

            return custom_filters, storage_filters

        return {}, filters

    def _process_custom_filters(self, result, filters):
        if not filters:
            return result
        for item in result[:]:
            for field_name, filter_value in filters.items():
                if not result:
                    break
                elif item not in result:
                    continue
                elif isinstance(filter_value, dm_filters.In):
                    if getattr(item, field_name) not in filter_value.value:
                        result.remove(item)
                        continue
                elif isinstance(filter_value, dm_filters.EQ):
                    if getattr(item, field_name) != filter_value.value:
                        result.remove(item)
                        continue
                else:
                    raise exc.ValidationFilterIncompatibleError(val=field_name)
        return result

    def _process_storage_filters(self, filters, order_by=None):
        return self.model.objects.get_all(filters=filters, order_by=order_by)

    @staticmethod
    def _convert_raw_filters_to_dm_filters(filters):
        """For use in places, where we manually work with input raw filters"""
        for k, v in filters.items():
            if not isinstance(v, dm_filters.AbstractClause):
                filters[k] = dm_filters.EQ(v)
        return filters

    def filter(self, filters, order_by=None):
        custom_filters, storage_filters = self._split_filters(filters)

        result = self._process_storage_filters(storage_filters, order_by=order_by)

        return self._process_custom_filters(result, custom_filters)

    def delete(self, uuid):
        self.get(uuid=uuid).delete()

    def update(self, uuid, **kwargs):
        dm = self.get(uuid=uuid)
        dm.update_dm(values=kwargs)
        dm.update()
        return dm


class BaseNestedResourceController(BaseResourceController):
    __pr_name__ = "parent_resource"

    def _prepare_kwargs(self, parent_resource, **kwargs):
        kw_params = kwargs.copy()
        kw_params[self.__pr_name__] = parent_resource
        return kw_params

    def create(self, **kwargs):
        return super(BaseNestedResourceController, self).create(
            **self._prepare_kwargs(**kwargs)
        )

    def get(self, **kwargs):
        return super(BaseNestedResourceController, self).get(
            **self._prepare_kwargs(**kwargs)
        )

    def filter(self, parent_resource, filters, order_by=None):
        filters = filters.copy()
        filters[self.__pr_name__] = dm_filters.EQ(parent_resource)
        return super(BaseNestedResourceController, self).filter(
            filters=filters, order_by=order_by
        )

    def delete(self, parent_resource, uuid):
        dm = self.get(parent_resource=parent_resource, uuid=uuid)
        dm.delete()

    def update(self, parent_resource, uuid, **kwargs):
        dm = self.get(parent_resource=parent_resource, uuid=uuid)
        dm.update_dm(values=kwargs)
        dm.update()
        return dm


class PaginationFilterBuilder:
    """
    Represents a pagination cursor with sort value and ID.
    Handles non-trivial logic for paginating by non-unique column.
    """

    def __init__(self, model, marker_id, sort_column=None, sort_direction="asc"):
        """
        Create pagination cursor from marker ID.

        Problem: when we sort by non-unique field, there'll be rows
        which were in the previous response too.
        So, we need to remove them somehow.
        Example of a target query:
            SELECT * FROM table WHERE
                filters
                AND
                (
                    order_key > marker_order_val
                        AND
                        (
                            order_key <> marker_order_val
                            OR
                            id_name > _pagination_marker
                        )
                    )
            ORDER BY
                order_key order_dir, id_name ASC
            LIMIT X;

        Args:
            model: The model to query
            marker_id: The ID of the last row from previous page
            sort_column: Column name to sort by (None for ID-only sort)
            sort_direction: "asc" or "desc"
        """
        self.marker_id = marker_id
        self.sort_col = sort_column
        self.sort_dir = sort_direction
        self.id_name = model.get_id_property_name()
        self.sort_value = self._fetch_sort_val(model)

    def _fetch_sort_val(self, model):
        """Fetch the sort value for this marker ID"""
        # No need for extra query, if we're sorting by UUID
        if self.sort_col in (None, self.id_name):
            return self.marker_id

        # Get last seen row
        marker_row = model.objects.get_one(
            filters={self.id_name: dm_filters.EQ(self.marker_id)}
        )
        return getattr(marker_row, self.sort_col)

    def build_filter(self):
        """Build the compound pagination filter."""
        if not self.sort_col:
            return {self.id_name: dm_filters.GT(self.marker_id)}

        # Check if sorting by ID directly
        id_op = dm_filters.GT if self.sort_dir == "asc" else dm_filters.LT
        if self.sort_col == self.id_name:
            return {self.id_name: id_op(self.marker_id)}

        # Standard compound cursor for non-ID primary sort
        return dm_filters.OR(
            {self.sort_col: id_op(self.sort_value)},
            dm_filters.AND(
                {self.sort_col: dm_filters.EQ(self.sort_value)},
                {self.id_name: dm_filters.GT(self.marker_id)},
            ),
        )


class BasePaginationMixin(object):
    """Pagination mixin, marker based, not offset based!

    Contract:
    - Works via `page_limit` and `page_marker` params.
    - All results are sorted by `order by id ASC`, if sort_key was not provided
    - page_limit: sets limit of each "page"
    - page_marker: sets marker of last ID in previous batch,
      next batch will by filtered by `where id > MARKER`
    - There are next "pages" while X-Pagination-Marker exists in response

    Example:
    # get first "page"
    curl '/some/collection/?page_limit=5' -i
    HTTP/1.1 200 OK
    X-Pagination-Limit: 5
    X-Pagination-Marker: XXX_UUID1

    # get next "page" by marker
    curl '/some/collection/?page_limit=5&page_marker=XXX_UUID1' -i
    HTTP/1.1 200 OK
    X-Pagination-Limit: 5
    X-Pagination-Marker: XXX_UUID2

    # get last "page" by marker
    curl '/some/collection/?page_limit=5&page_marker=XXX_UUID2' -i
    HTTP/1.1 200 OK
    X-Pagination-Limit: 5
    (Last page won't have Marker)
    """

    _pagination_limit = 0
    _header_page_limit = "X-Pagination-Limit"
    _header_page_marker = "X-Pagination-Marker"

    _param_page_limit = "page_limit"
    _param_page_marker = "page_marker"

    def _create_response(self, body, status, headers):
        if self._pagination_limit:
            headers[self._header_page_limit] = str(self._pagination_limit)
            if len(body) == self._pagination_limit:
                headers[self._header_page_marker] = str(
                    getattr(body[-1], self.model.get_id_property_name())
                )

        return super(BasePaginationMixin, self)._create_response(body, status, headers)

    def _prepare_pagination_meta(self):
        try:
            self._pagination_limit = int(
                self._req.api_context.params.get(self._param_page_limit, 0)
            )
            if self._pagination_limit < 0:
                raise ValueError()
        except ValueError:
            raise exc.ParseError(value="%s" % (self._pagination_limit,))
        # TODO(g.melikov): do we need to validate if marker ID record exists?
        self._pagination_marker = self._req.api_context.params.get(
            self._param_page_marker
        )
        if self._pagination_marker:
            self._pagination_marker = (
                self._parse_resource_uuid(
                    "uuid",
                    self._pagination_marker,
                    self.get_resource().get_id_type(),
                )
                if self.__resource__
                else self._pagination_marker
            )

    def do_collection(self, parent_resource=None):
        self._prepare_pagination_meta()

        return super(BasePaginationMixin, self).do_collection(
            parent_resource=parent_resource
        )

    def _process_storage_filters(self, filters, order_by=None):
        self._validate_params(filters, order_by)
        filters, order_by = self._build_pagination_with_cursor(filters, order_by)
        return self.model.objects.get_all(
            filters=filters,
            limit=self._pagination_limit,
            order_by=order_by,
        )

    def _validate_params(self, filters, order_by):
        if order_by and len(order_by) > 1:
            raise exc.ValidationSortNumberError()

    def _build_pagination_with_cursor(self, filters, order_by):
        if self._pagination_marker:
            sort_col, sort_dir = (
                next(iter(order_by.items())) if order_by else (None, "asc")
            )
            cursor = PaginationFilterBuilder(
                self.model,
                self._pagination_marker,
                sort_col,
                sort_dir,
            )
            pagination_filters = cursor.build_filter()
            filters = dm_filters.AND(pagination_filters, filters)

        # Build final ordering
        # Don't add ID tiebreaker if ID is already the primary sort
        id_name = self.model.get_id_property_name()
        if order_by:
            order_by = order_by.copy()
            if id_name not in order_by:
                order_by[id_name] = "asc"
        else:
            order_by = {id_name: "asc"}

        return filters, order_by

    def paginated_filter(self, filters, order_by=None):
        custom_filters, storage_filters = self._split_filters(filters)

        cleaned_results = []

        # Get additional data from DB if some was filtered by custom props
        while len(cleaned_results) < self._pagination_limit:
            result = self._process_storage_filters(storage_filters, order_by=order_by)

            if not len(result):
                break

            self._pagination_marker = getattr(
                result[-1], self.model.get_id_property_name()
            )

            cleaned_results.extend(self._process_custom_filters(result, custom_filters))

        if len(cleaned_results) > self._pagination_limit:
            cleaned_results = cleaned_results[: self._pagination_limit]

        return cleaned_results


class BaseResourceControllerPaginated(BasePaginationMixin, BaseResourceController):
    def filter(self, filters, order_by=None):
        # NOTE(g.melikov): if you want to add pagination and you need to
        #  override `filter` method - it's better to add custom filters into
        #  `_process_custom_filters`.
        if self._pagination_limit:
            return self.paginated_filter(filters, order_by=order_by)

        return super(BasePaginationMixin, self).filter(filters, order_by=order_by)


class BaseNestedResourceControllerPaginated(
    BasePaginationMixin, BaseNestedResourceController
):
    def filter(self, parent_resource, filters, order_by=None):
        filters = filters.copy()
        filters[self.__pr_name__] = dm_filters.EQ(parent_resource)

        if self._pagination_limit:
            return self.paginated_filter(filters, order_by=order_by)

        return super(BaseNestedResourceController, self).filter(
            filters, order_by=order_by
        )


class RoutesListController(Controller):
    __TARGET_PATH__ = "/"

    def _get_target_route(self, target_path):
        """
        Finds the route from main route by target_path.

        :param target_path: the path to find the route from main route
        :return: the target route
        """
        result = self.request.application.main_route
        for next_path in target_path.rstrip("/").split("/")[1:]:
            result = getattr(result, next_path)
        return result

    def filter(self, filters, order_by=None):
        """
        Returns a list of all routes in the main route that are collection
        routes.

        The `__TARGET_PATH__` attribute on the class determines the target
        route that is filtered.

        :param filters: the filters to apply when filtering the routes (Is not
            implemented now)
        :param order_by: fields to order the routes by (Is not implemented
            now)
        :return: a list of route names
        """
        target_route = self._get_target_route(self.__TARGET_PATH__)
        req = self.request
        return [
            route_name
            for route_name in target_route.get_routes()
            if target_route.get_route(route_name)(req).is_collection_route()
        ]


class RootController(RoutesListController):
    def filter(self, filters, order_by=None):
        main_route = self.request.application.main_route
        req = self.request
        return [
            route_name
            for route_name in main_route.get_routes()
            if main_route.get_route(route_name)(req).is_collection_route()
        ]


class OpenApiSpecificationController(Controller):
    @oa_utils.extend_schema(
        summary="OpenApi specification",
        responses=oa_c.build_openapi_object_response(
            properties={}, description="OpenApi specification"
        ),
        operation_id="Get_OpenApi_specification",
    )
    def get(self, uuid):
        openapi_engine = self.request.application.openapi_engine
        if openapi_engine:
            return openapi_engine.build_openapi_specification(
                version=uuid,
                request=self._req,
            )
        raise exc.NotExtended()

    def filter(self, filters, order_by=None):
        openapi_engine = self.request.application.openapi_engine
        if openapi_engine:
            return openapi_engine.list_supported_openapi_versions()
        return []



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from restalchemy.common import exceptions


class RecordNotFound(exceptions.RestAlchemyException):
    code = 404
    message = (
        "Can't found record in storage for model (%(model)s) and filters (%(filters)s)."
    )


class MultipleUpdatesDetected(exceptions.RestAlchemyException):
    message = "Multiple records were updated in storage for %(model)s"


class HasManyRecords(exceptions.RestAlchemyException):
    message = (
        "Has many records in storage for model (%(model)s) and filters (%(filters)s)."
    )


class DeadLock(exceptions.RestAlchemyException):
    message = "Deadlock found when trying to get lock. Original message: %(msg)s"


class ConflictRecords(exceptions.RestAlchemyException):
    code = 409
    message = "Duplicate parameters for '%(model)s'. Original message: %(msg)s"

    def __init__(self, **kwargs):
        super(ConflictRecords, self).__init__(**kwargs)
        self._original_msg = kwargs.get("msg") or ""

    def _re_parser(self, re_template):
        result = re.search(re_template, self._original_msg)
        if result is None:
            raise ValueError(
                f"Incorrect message for parsing. {self._original_msg}"
                f" but should be {re_template}"
            )
        return result.groups()

    @property
    def _parsed_message_mysql(self):
        re_template = r"Duplicate entry '(.*)' for key '(.*)'"
        result = self._re_parser(re_template)
        return {
            "key": result[1],
            "value": result[0],
        }

    @property
    def _parsed_message_postgresql(self):
        re_template = r"DETAIL:\s+Key \((\w+)\)=\(([^)]+)\)"
        result = self._re_parser(re_template)
        return {
            "key": result[0],
            "value": result[1],
        }

    @property
    def _parse_message(self):
        if self._original_msg.startswith("Duplicate entry"):
            return self._parsed_message_mysql
        elif self._original_msg.startswith("duplicate key value violates"):
            return self._parsed_message_postgresql
        raise ValueError(f"Can't parse message: {self._original_msg}")

    @property
    def value(self):
        return self._parse_message["value"]

    @property
    def key(self):
        return self._parse_message["key"]


class UnknownStorageException(exceptions.RestAlchemyException):
    message = "Unknown storage exception: %(caused)r"

    def __init__(self, caused, **kwargs):
        self._caused = caused
        super(UnknownStorageException, self).__init__(caused=caused, **kwargs)

    @property
    def caused(self):
        return self._caused



# Copyright 2021 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.api.middlewares.logging import LoggingMiddleware
from restalchemy.tests.unit import base


class LoggingMiddlewareTestCase(base.BaseTestCase):
    def test_sanitize_authorization_header(self):
        request_mock = mock.MagicMock()
        request_mock.headers = {
            "fake_header": "fake_header_value",
            "Authorization": "basic something",
        }
        middlew = LoggingMiddleware("application")
        checks = {}

        def _check_sanitized_header(msg, *args, **kwargs):
            if "API > " in msg:  # test only request logs
                checks["check_run"] = True
                for header in args[1]:
                    if header.startswith("Authorization"):
                        if "something" not in header:
                            checks["Authorization"] = True
                    if header.startswith("fake_header"):
                        if "fake_header_value" in header:
                            checks["fake_header"] = True

        middlew.logger.debug = mock.Mock(side_effect=_check_sanitized_header)

        middlew.process_request(request_mock)

        self.assertTrue(checks["check_run"])
        self.assertTrue(checks["Authorization"])
        self.assertTrue(checks["fake_header"])
        self.assertDictEqual(
            {
                "fake_header": "fake_header_value",
                "Authorization": "basic something",
            },
            request_mock.headers,
        )



#    Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#    Copyright 2021 Eugene Frolov.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import logging
import orjson
import types

from restalchemy.api import constants
from restalchemy.common import exceptions
from restalchemy.common import utils

DEFAULT_VALUE = object()
CONTENT_TYPE_APPLICATION_JSON = DEFAULT_CONTENT_TYPE = (
    constants.CONTENT_TYPE_APPLICATION_JSON
)  # noqa

LOG = logging.getLogger(__name__)


def get_content_type(headers):
    return headers.get("Content-Type") or constants.DEFAULT_CONTENT_TYPE


class BaseResourcePacker(object):
    _skip_none = True

    def __init__(self, resource_type, request):
        self._rt = resource_type
        self._req = request

    def pack_resource(self, obj):
        if isinstance(
            obj,
            (str, int, float, bool, type(None), list, tuple, dict),
        ):
            return obj
        else:
            result = {}
            for name, prop in self._rt.get_fields_by_request(self._req):
                api_name = prop.api_name
                if prop.is_public() and not self._rt._fields_permissions.is_hidden(
                    name, self._req
                ):
                    value = getattr(obj, name)
                    if value is None:
                        if not self._skip_none:
                            result[api_name] = value
                    else:
                        result[api_name] = prop.dump_value(value)

            return result

    def pack(self, obj):
        if isinstance(obj, list) or isinstance(obj, types.GeneratorType):
            return [self.pack_resource(resource) for resource in obj]
        else:
            return self.pack_resource(obj)

    @utils.raise_parse_error_on_fail
    def _parse_value(self, name, value, prop):
        return prop.parse_value(self._req, value)

    def unpack(self, value):
        if not self._rt:
            return value
        value = copy.deepcopy(value)
        result = {}
        for name, prop in self._rt.get_fields_by_request(self._req):
            api_name = prop.api_name
            prop_value = value.pop(api_name, DEFAULT_VALUE)
            if prop_value is not DEFAULT_VALUE:
                if not prop.is_public():
                    raise exceptions.ValidationPropertyPrivateError(property=api_name)

                if self._rt._fields_permissions.is_readonly(name, self._req):
                    raise exceptions.FieldPermissionError(field=name)
                result[name] = self._parse_value(api_name, prop_value, prop)

        if len(value) > 0:
            raise exceptions.ValidationPropertyIncompatibleError(
                val=value, model=self._rt.get_model().__name__
            )

        return result


class JSONPacker(BaseResourcePacker):
    def pack(self, obj):
        return orjson.dumps(
            super(JSONPacker, self).pack(obj), option=orjson.OPT_NON_STR_KEYS
        )

    def unpack(self, value):
        if isinstance(value, bytes):
            try:
                return super(JSONPacker, self).unpack(
                    orjson.loads(value),
                )
            except orjson.JSONDecodeError:
                raise exceptions.ParseBodyError()

        return super(JSONPacker, self).unpack(orjson.loads(value))


class JSONPackerIncludeNullFields(JSONPacker):
    _skip_none = False


class MultipartPacker(JSONPacker):
    """
    This packer is specifically designed to handle multipart/form-data
    requests, which are commonly used for uploading files.
    It ensures that only file-related data is extracted and processed,
    allowing for seamless integration with file upload or download function.

    Key Features:
    - Supports only file uploads by default.
    - If future requirements extend the support to include model JSON under
      a specific key, it may be updated accordingly.
    - Extracts file data into a structured dictionary where each file is
      identified by its part name.
    - Uses '_multipart' as a flag to indicate that multipart content was used
      for processing.
    """

    # TODO: as of now this packer support only file uploads.
    #  In future model's json should be under this field name
    # _resource_key = "_resource"
    _multipart_key = "multipart"
    _parts_key = "parts"

    def _unpack_multipart(self):
        """
        Unpacks multipart/form-data request into a structured dictionary.

        :return: A dictionary containing the following keys:
            - 'multipart': A boolean flag indicating that multipart content
              was found.
            - 'parts': A dictionary where each key is the part field name and
              the corresponding value is the file data (bytes)
              (as `FieldStorage()`).
        """
        result = collections.defaultdict(dict)
        result[self._multipart_key] = True

        # if self._resource_key not in self._req.POST:
        #     ValueError("Resource data should be under '_resource' part!")
        #     result[self._resource_key] = super().unpack(self._req.POST['self._resource_key'])

        for key, part in self._req.POST.items():
            # if key == self._resource_key:
            #     continue
            result[self._parts_key][key] = part

        return result

    def unpack(self, value):
        if constants.CONTENT_TYPE_MULTIPART in self._req.content_type:
            return self._unpack_multipart()
        return super().unpack(value)

    def pack(self, obj):
        if isinstance(obj, bytes):
            return obj
        return super().pack(obj)


packer_mapping = {
    constants.CONTENT_TYPE_APPLICATION_JSON: JSONPacker,
    constants.CONTENT_TYPE_MULTIPART: MultipartPacker,
}


def parse_content_type(value):
    # Cleanup: application/json;charset=UTF-8
    return value.split(";")[0].strip() if value else None


def get_packer(content_type):
    try:
        return packer_mapping[parse_content_type(content_type)]
    except KeyError:
        # TODO(Eugene Frolov): Specify Exception Type and message
        raise Exception("Packer can't found for content type %s " % content_type)


def set_packer(content_type, packer_class):
    packer_mapping[content_type] = packer_class



# Copyright 2022 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import gc

from restalchemy.common import utils
from restalchemy.storage.sql import engines
from restalchemy.tests.functional import consts


class DBEngineMixin(object):
    __ENGINE__ = None

    @utils.classproperty
    def engine(cls):
        return cls.__ENGINE__

    @classmethod
    def init_engine(cls):
        engines.engine_factory.configure_factory(db_url=consts.get_database_uri())
        cls.__ENGINE__ = engines.engine_factory.get_engine()

    @classmethod
    def destroy_engine(cls):
        # Note(efrolov): Must be deleted otherwise we will start collect
        #                connections and get an error "too many connections"
        #                from MySQL
        del cls.__ENGINE__
        engines.engine_factory.destroy_engine()
        # Force GC collection to free database active connections
        gc.collect()

    @classmethod
    def get_all_tables(cls, session=None):
        with cls.engine.session_manager(session=session) as s:
            if session.engine.dialect.name == "mysql":
                res = s.execute("""
                    select
                        table_name as table_name
                    from information_schema.tables
                    where table_schema = database();
                """).fetchall()
            elif session.engine.dialect.name == "postgresql":
                res = s.execute("""
                    select
                        table_name as table_name
                    from information_schema.tables
                    where table_schema = current_schema();
                """).fetchall()
            else:
                raise NotImplementedError("Unsupported dialect")
        tables = {row["table_name"] for row in res}
        return tables

    @classmethod
    def is_table_exists(cls, table_name, session=None):
        with cls.engine.session_manager(session=session) as s:
            tables = cls.get_all_tables(session=s)
        res = table_name in tables
        return res

    @classmethod
    def drop_table(cls, table_name, session=None, cascade=False):
        cascade = " CASCADE" if cascade else ""
        with cls.engine.session_manager(session=session) as s:
            s.execute(
                f"drop table if exists {session.engine.escape(table_name)}{cascade}"
            )

    @classmethod
    def truncate_table(cls, table_name, session=None):
        with cls.engine.session_manager(session=session) as s:
            s.execute(f"truncate table {session.engine.escape(table_name)}")

    @classmethod
    def drop_all_tables(cls, session=None, cascade=False):
        with cls.engine.session_manager(session=session) as s:
            tables = cls.get_all_tables(session=s)
            for table in tables:
                cls.drop_table(table, session=s, cascade=cascade)

    @classmethod
    def get_all_views(cls, session=None):
        with cls.engine.session_manager(session=session) as s:
            if session.engine.dialect.name == "mysql":
                res = s.execute("""
                    select
                        table_name as table_name
                    from information_schema.views
                    where table_schema = database();
                """).fetchall()
            elif session.engine.dialect.name == "postgresql":
                res = s.execute("""
                    select
                        table_name as table_name
                    from information_schema.views
                    where table_schema = current_schema();
                """).fetchall()
            else:
                raise NotImplementedError("Unsupported dialect")
        return {row["table_name"] for row in res}

    @classmethod
    def drop_all_views(cls, session=None):
        with cls.engine.session_manager(session=session) as s:
            views = cls.get_all_views(session=s)
            for view in views:
                cls.drop_view(view, session=s)

    @classmethod
    def drop_view(cls, view_name, session=None):
        with cls.engine.session_manager(session=session) as s:
            s.execute(f"drop view if exists {session.engine.escape(view_name)}")



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import logging
import threading

from restalchemy.common import exceptions as e
from restalchemy.storage.sql import engines

LOG = logging.getLogger(__name__)


class Context(object):
    def __init__(self, engine_name=engines.DEFAULT_NAME):
        """
        Initializes the context with a given engine name.

        :param engine_name: The name of the engine to use. Defaults to
            DEFAULT_NAME.
        :type engine_name: str
        """
        super(Context, self).__init__()
        self._engine_name = engine_name

    def start_new_session(self):
        """
        Starts a new session.

        :return: A session object.
        :rtype: Session
        """
        engine = self._engine
        storage = engine.get_session_storage()
        session = engine.get_session()
        try:
            storage.store_session(session)
        except Exception:
            # Return the session to the pool and raise the exception.
            session.close()
            raise
        LOG.debug("New session %r has been started", session)
        return session

    @property
    def _engine(self):
        """
        Property that returns the current engine instance.

        The engine instance is retrieved from the engine factory based on the
        engine name provided during object initialization. If the engine
        instance does not exist, a ValueError is raised.

        :returns: The current engine instance.
        :rtype: AbstractEngine
        :raises ValueError: If the engine instance does not exist.
        """
        return engines.engine_factory.get_engine(name=self._engine_name)

    @contextlib.contextmanager
    def session_manager(self):
        """
        A context manager for managing a database session.

        This method provides a context for executing database operations within
        a session. If a session is not provided, it attempts to retrieve one
        from storage or creates a new session. The session is automatically
        committed if no exceptions occur, otherwise it is rolled back. The
        session is always closed when the context exits.

        :returns: The active session for database operations.
        :raises Exception: Any exceptions raised during the session's execution
            will result in a rollback of the session.
        """
        session = self.start_new_session()
        try:
            yield session
            session.commit()
            LOG.debug("Session %r has been committed", session)
        except Exception:
            session.rollback()
            LOG.exception("Session %r has been rolled back by reason:", session)
            raise
        finally:
            self.session_close()

    def _get_storage(self):
        """
        Retrieve the session storage object from the engine.

        This method accesses the engine associated with the context to
        obtain the session storage object. The session storage object
        is responsible for managing session objects within the engine.

        :returns: The session storage object.
        :rtype: SessionThreadStorage
        """

        engine = self._engine
        return engine.get_session_storage()

    def get_session(self):
        """
        Returns a session object from the session storage associated with
        this context's engine.

        If a session object does not exist in the session storage, a new
        session object is created and returned.

        :returns: A session object.
        :rtype: Session
        """
        return self._get_storage().get_session()

    def session_close(self):
        """
        Close a session object from thread storage and remove it from
        thread storage.

        If session object is not found in thread storage, this method
        doesn't do anything.

        :raises: Any exceptions raised during session's closing will
            result in a log message and then discarded.
        """
        session = self.get_session()
        try:
            session.close()
            LOG.debug("Session %r has been closed", session)
        except Exception:
            LOG.exception("Can't close session by reason:")
        finally:
            self._get_storage().remove_session()
            LOG.debug("Session %r has been removed from thread storage", session)


class StorageRuntimeException(e.RestAlchemyException):
    __template__ = "Storage runtime error: %(message)s"


class ReadOnlyStorage(StorageRuntimeException):
    __template__ = "%(name)s Key is read only."


class ContextRuntimeException(e.RestAlchemyException):
    __template__ = "Context runtime error: %(message)s"


class ContextAlreadyInStorage(ContextRuntimeException):
    def __init__(self, message="Context is already in storage."):
        super().__init__(message=message)


class ContextIsNotExistsInStorage(ContextRuntimeException):
    def __init__(self, message="Context is not exists in storage."):
        super().__init__(message=message)


class Storage:
    def __init__(self, data: dict = None):
        """
        Initialize the global storage.

        :param data: The dictionary with initial data.
        :type data: dict
        """
        self._storage = {}
        if data is not None:
            self._storage = data.copy()

    def put(self, name, value, read_only=False):
        """
        Store a key-value pair in the storage.

        :param name: The name of the key to store the value.
        :type name: str
        :param value: The value to be stored associated with the key.
        :type value: object
        :param read_only: A boolean indicating whether the key is read only.
            If True, the key can't be changed once it is set.
        :type read_only: bool
        """

        if name in self._storage and self._storage[name]["read_only"]:
            raise ReadOnlyStorage(name=name)

        self._storage[name] = {
            value: value,
            "read_only": read_only,
        }

    def get(self, name):
        """
        Get the value from the storage by name.

        :param name: The name of the key in the storage.
        :type name: str
        :return: The value associated with the key.
        :rtype: object
        :raises KeyError: When there is no key with the given name.
        """
        return self._storage[name]["value"]

    def delete(self, name, force=False):
        """
        Delete the key-value pair from the storage.

        :param name: The name of the key to delete.
        :type name: str
        :param force: Force deletion of the key-value pair even if
            the key is read only.
        :type force: bool
        :raises KeyError: If there is no key with the given name.
        :raises ReadOnlyStorageError: If the key is read only and
            force is False.
        """

        if name in self._storage and self._storage[name]["read_only"]:
            if not force:
                raise ReadOnlyStorage(name=name)

        del self._storage[name]


class ContextWithStorage(Context):
    _local_thread_storage = threading.local()

    def __init__(
        self,
        engine_name: str = engines.DEFAULT_NAME,
        context_storage: Storage = None,
    ):
        """
        Initialize the context with storage.

        :param engine_name: The name of the engine to use. Defaults to
            DEFAULT_NAME.
        :type engine_name: str
        :param context_storage: The storage object to use. Defaults to None,
            which means a new Storage object will be created.
        :type context_storage: Storage
        """
        super(ContextWithStorage, self).__init__(engine_name=engine_name)
        self._context_storage = context_storage or Storage()

    @property
    def context_storage(self):
        """
        The storage object associated with the context.

        :rtype: Storage
        """
        return self._context_storage

    @classmethod
    def _store_context_session(cls, context):
        """
        Store the context in the local thread storage.

        :param context: The context to be stored.
        :type context: ContextWithStorage
        :raises RuntimeError: If the context is already stored.
        """
        if hasattr(cls._local_thread_storage, "context"):
            raise ContextAlreadyInStorage()

        cls._local_thread_storage.context = context

    @classmethod
    def _clear_context(cls):
        """
        Clear the context from the local thread storage.

        :raises ContextIsNotExistsInStorage: If there is no context associated
            with the local thread.
        """
        if not hasattr(cls._local_thread_storage, "context"):
            raise ContextIsNotExistsInStorage()
        del cls._local_thread_storage.context

    @classmethod
    def get_context(cls):
        """
        Get the current context.

        Returns the current context from the storage associated with the local
        thread.

        :return: The current context.
        :rtype: ContextWithStorage
        :raises ContextIsNotExistsInStorage: If there is no context associated
            with the local thread.
        """
        if not hasattr(cls._local_thread_storage, "context"):
            raise ContextIsNotExistsInStorage()
        return cls._local_thread_storage.context

    @contextlib.contextmanager
    def context_manager(self):
        """
        Context manager to manage the lifecycle of a context.

        This context manager temporarily stores the context in a local thread
        storage, making it accessible within the managed block. Upon exiting
        the block, the context is cleared from the storage.

        :yields: The current context.
        :raises ContextAlreadyInStorage: If a context is already stored.
        :raises ContextIsNotExistsInStorage: If the context does not exist in
            storage when attempting to clear it.
        """

        LOG.debug("Start context manager with context %r", self)
        self._store_context_session(self)
        try:
            yield self.get_context()
        finally:
            LOG.debug("End context manager with context %r", self)
            self._clear_context()


def get_context():
    """
    Get the current context.

    Returns the current context from the storage associated with the local
    thread.

    :return: The current context.
    :rtype: ContextWithStorage
    :raises ContextIsNotExistsInStorage: If there is no context associated
        with the local thread.
    """
    return ContextWithStorage.get_context()



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage import exceptions as exc
from restalchemy.storage.sql import orm
from restalchemy.tests.functional import base


class BatchInsertModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "batch_insert"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


class InsertTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    def test_correct_batch_insert(self):
        model1 = BatchInsertModel(foo_field1=1, foo_field2="Model1")
        model2 = BatchInsertModel(foo_field1=2, foo_field2="Model2")
        model3 = BatchInsertModel(foo_field1=3, foo_field2="Model3")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2, model3])

        all_models = set(BatchInsertModel.objects.get_all())

        self.assertEqual({model1, model2, model3}, all_models)

    def test_duplicate_primary_key_batch_insert(self):
        dup_uuid = uuid.uuid4()
        model1 = BatchInsertModel(uuid=dup_uuid, foo_field1=1, foo_field2="Model1")
        model2 = BatchInsertModel(foo_field1=2, foo_field2="Model2")
        model3 = BatchInsertModel(uuid=dup_uuid, foo_field1=3, foo_field2="Model3")
        # NOTE(efrolov): PRIMARY - is value from table structure,
        #   unique index for any primary key. Constant in mysql database.
        key_name = "PRIMARY" if self.engine.dialect.name == "mysql" else "uuid"

        with self.engine.session_manager() as session:
            with self.assertRaises(exc.ConflictRecords):
                try:
                    session.batch_insert([model1, model2, model3])
                except exc.ConflictRecords as e:
                    self.assertEqual(key_name, e.key)
                    # NOTE(efrolov): all values from exception in string type
                    self.assertEqual(str(dup_uuid), e.value)
                    raise

        all_models = BatchInsertModel.objects.get_all()

        self.assertEqual([], all_models)

    def test_duplicate_secondary_key_batch_insert(self):
        dup_value = 2
        model1 = BatchInsertModel(foo_field1=1, foo_field2="Model1")
        model2 = BatchInsertModel(foo_field1=dup_value, foo_field2="Model2")
        model3 = BatchInsertModel(foo_field1=dup_value, foo_field2="Model3")

        with self.engine.session_manager() as session:
            with self.assertRaises(exc.ConflictRecords):
                try:
                    session.batch_insert([model1, model2, model3])
                except exc.ConflictRecords as e:
                    self.assertEqual("foo_field1", e.key)
                    # NOTE(efrolov): all values from exception in string type
                    self.assertEqual(str(dup_value), e.value)
                    raise

        all_models = BatchInsertModel.objects.get_all()

        self.assertEqual([], all_models)



# Copyright 2022 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import mock

from restalchemy.api.middlewares import logging_http
from restalchemy.common import exceptions
from restalchemy.tests.unit import base

from webob import request


class LoggingHttpMiddlewareTestCase(base.BaseTestCase):
    middlew = logging_http.LoggingHttpMiddleware("application")

    def test_response_with_exception(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = exceptions.RestAlchemyException

        exc = exceptions.RestAlchemyException

        with mock.patch.object(logging.Logger, "info") as log_info:
            with self.assertRaises(exc):
                self.middlew.process_request(request_mock)
            log_info.assert_called()

    def test_response(self):
        res = logging_http.Http500FakeResponse(status_code=200)

        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.return_value = res

        with mock.patch.object(logging.Logger, "info") as log_info:
            self.middlew.process_request(request_mock)
            log_info.assert_called()



#    Copyright 2021 Eugene Frolov <eugene@frolov.net.ru>.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.storage.sql import orm

# root table
FAKE_ROOT_INT = 0
FAKE_ROOT_STR = "root"
FAKE_ROOT_BOOL = True

# lnp1_1 table
FAKE_LNP1_1_INT = 11
FAKE_LNP1_1_STR = "lnp1_1"
FAKE_LNP1_1_BOOL = True

# lnp1_2 table
FAKE_LNP1_2_INT = 12
FAKE_LNP1_2_STR = "lnp1_2"
FAKE_LNP1_2_BOOL = True

# lwp1_1 table
FAKE_LWP1_1_INT = 111
FAKE_LWP1_1_STR = "lwp1_1"
FAKE_LWP1_1_BOOL = True

# lwp1_2 table
FAKE_LWP1_2_INT = 112
FAKE_LWP1_2_STR = "lwp1_2"
FAKE_LWP1_2_BOOL = True

# lwp2_1 table
FAKE_LWP2_1_INT = 21
FAKE_LWP2_1_STR = "lwp2_1"
FAKE_LWP2_1_BOOL = True

# lwp2_2 table
FAKE_LWP2_2_INT = 22
FAKE_LWP2_2_STR = "lwp2_2"
FAKE_LWP2_2_BOOL = True

# lwp2_3 table
FAKE_LWP2_3_INT = 23
FAKE_LWP2_3_STR = "lwp2_3"
FAKE_LWP2_3_BOOL = True

# lwp2_4 table
FAKE_LWP2_4_INT = 24
FAKE_LWP2_4_STR = "lwp2_4"
FAKE_LWP2_4_BOOL = True


# TODO(efrolov): convert to mock wrapper for handle all methods
OBJECT_COLLECTION_MOCK = mock.Mock(spec=orm.ObjectCollection)


class ObjectCollection(orm.ObjectCollection):
    def get_all(
        self,
        filters=None,
        session=None,
        cache=False,
        limit=None,
        order_by=None,
        locked=False,
    ):
        OBJECT_COLLECTION_MOCK.get_all(
            filters=filters,
            session=session,
            cache=cache,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )
        return super(ObjectCollection, self).get_all(
            filters=filters,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )


class SQLStorableMixin(orm.SQLStorableMixin):
    _ObjectCollection = ObjectCollection


class Root(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "root"

    field_int = properties.property(types.Integer(), default=FAKE_ROOT_INT)
    field_str = properties.property(types.String(), default=FAKE_ROOT_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_ROOT_BOOL)


class LNP1_1(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lnp1_1"

    root = relationships.relationship(Root, required=True, prefetch=False)
    field_int = properties.property(types.Integer(), default=FAKE_LNP1_1_INT)
    field_str = properties.property(types.String(), default=FAKE_LNP1_1_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LNP1_1_BOOL)


class LNP1_2(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lnp1_2"

    root = relationships.relationship(Root, required=False, prefetch=False)
    field_int = properties.property(types.Integer(), default=FAKE_LNP1_2_INT)
    field_str = properties.property(types.String(), default=FAKE_LNP1_2_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LNP1_2_BOOL)


class LWP1_1(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lwp1_1"

    root = relationships.relationship(Root, required=True, prefetch=True)
    field_int = properties.property(types.Integer(), default=FAKE_LWP1_1_INT)
    field_str = properties.property(types.String(), default=FAKE_LWP1_1_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LWP1_1_BOOL)


class LWP1_2(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lwp1_2"

    root = relationships.relationship(Root, required=False, prefetch=True)
    field_int = properties.property(types.Integer(), default=FAKE_LWP1_2_INT)
    field_str = properties.property(types.String(), default=FAKE_LWP1_2_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LWP1_2_BOOL)


class LWP2_1(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lwp2_1"

    lwp1_1 = relationships.relationship(LWP1_1, required=True, prefetch=True)
    field_int = properties.property(types.Integer(), default=FAKE_LWP2_1_INT)
    field_str = properties.property(types.String(), default=FAKE_LWP2_1_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LWP2_1_BOOL)


class LWP2_2(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lwp2_2"

    lwp1_1 = relationships.relationship(LWP1_1, required=False, prefetch=True)
    field_int = properties.property(types.Integer(), default=FAKE_LWP2_2_INT)
    field_str = properties.property(types.String(), default=FAKE_LWP2_2_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LWP2_2_BOOL)


class LWP2_3(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lwp2_3"

    lwp1_2 = relationships.relationship(LWP1_2, required=True, prefetch=True)
    field_int = properties.property(types.Integer(), default=FAKE_LWP2_3_INT)
    field_str = properties.property(types.String(), default=FAKE_LWP2_3_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LWP2_3_BOOL)


class LWP2_4(models.ModelWithUUID, SQLStorableMixin):
    __tablename__ = "lwp2_4"

    lwp1_2 = relationships.relationship(LWP1_2, required=False, prefetch=True)
    field_int = properties.property(types.Integer(), default=FAKE_LWP2_4_INT)
    field_str = properties.property(types.String(), default=FAKE_LWP2_4_STR)
    field_bool = properties.property(types.Boolean(), default=FAKE_LWP2_4_BOOL)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.dm import filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import orm


# CREATE TABLE `foos` (
#      `uuid` CHAR(36) CHARACTER SET utf8 COLLATE utf8_unicode_ci NOT NULL ,
#      `foo_field1` INT NOT NULL ,
#      `foo_field2` VARCHAR(255) NOT NULL ,
# PRIMARY KEY (`uuid`(36)) USING HASH)
# ENGINE = InnoDB;
class FooModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "foos"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


# CREATE TABLE `test`.`bars` (
#     `uuid` CHAR(36) CHARACTER SET utf32 COLLATE utf32_unicode_ci NOT NULL ,
#     `bar_field1` VARCHAR(10) CHARACTER SET utf8 COLLATE utf8_unicode_ci
#         NOT NULL ,
#     `foo` CHAR(36) CHARACTER SET utf32 COLLATE utf32_unicode_ci NOT NULL ,
# PRIMARY KEY (`uuid`(36)) USING HASH,
# INDEX `rlfoo` (`foo`(36)) USING HASH)
# ENGINE = InnoDB;


# ALTER TABLE `bars` ADD CONSTRAINT `_idx_foo` FOREIGN KEY (`foo`)
# REFERENCES `foos`(`uuid`) ON DELETE RESTRICT ON UPDATE RESTRICT;
class BarModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "bars"
    bar_field1 = properties.property(types.String(min_length=1, max_length=10))
    foo = relationships.relationship(FooModel)


engines.engine_factory.configure_factory(db_url="mysql://test:test@127.0.0.1/test")


# Create new foo object and store it
foo1 = FooModel(foo_field1=10)
foo1.save()

bar1 = BarModel(bar_field1="test", foo=foo1)
bar1.save()

print(list(BarModel.objects.get_all()))
print(BarModel.objects.get_one(filters={"uuid": bar1.get_id()}))
print(list(BarModel.objects.get_all(filters={"foo": foo1})))
print(bar1.as_plain_dict())

bar1.delete()

foo2 = FooModel(foo_field1=11, foo_field2="some text")
foo2.save()

foos = list(FooModel.objects.get_all())
print(foos)

print(FooModel.objects.get_one(filters={"foo_field1": filters.EQ(10)}))

# Modify foo_field2 and update it in storage
foo2.foo_field2 = "xxx2 asdad asdasd"
foo2.save()

# Delete foo object from storage
for foo in foos:
    foo.delete()

print("foo_field1 is greater than 5")
for num in range(10):
    foo = FooModel(foo_field1=num)
    foo.save()

print(list(FooModel.objects.get_all(filters={"foo_field1": filters.GT(5)})))

print("foo_field1 is equal to 5 or 6")
print(FooModel.objects.get_all(filters={"foo_field1": filters.In([5, 6])}))

for model in FooModel.objects.get_all():
    model.delete()

# Complex filters
# WHERE ((`name1` = 1 AND `name2` = 2) OR (`name2` = 3))
filter_list = filters.OR(
    filters.AND({"name1": filters.EQ(1), "name2": filters.EQ(2)}),
    filters.AND({"name2": filters.EQ(3)}),
)

print(FooModel.objects.get_one(filters=filter_list))



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
from wsgiref import simple_server

from restalchemy.api import applications
from restalchemy.api.middlewares import (
    retry_on_error as retry_error_middleware,
)

from restalchemy.api.middlewares import errors as errors_middleware
from restalchemy.openapi import engines as openapi_engines
from restalchemy.openapi import structures as openapi_structures
from restalchemy.storage import exceptions as storage_exc
from restalchemy.storage.sql import engines
from restalchemy.tests.functional import consts
from restalchemy.tests.functional.restapi.ra_based.microservice import (
    middlewares,
)
from restalchemy.tests.functional.restapi.ra_based.microservice import routes


class RESTService(threading.Thread):
    def __init__(self, bind_host, bind_port, app_root):
        super(RESTService, self).__init__(name="REST Service")

        self._service_port = bind_port
        self._service_host = bind_host

        self._httpd = simple_server.make_server(
            bind_host,
            bind_port,
            app_root,
            simple_server.WSGIServer,
        )

    @property
    def service_port(self):
        return self._service_port

    @property
    def service_host(self):
        return self._service_host

    def run(self):
        self._httpd.serve_forever()

    def stop(self):
        self._httpd.server_close()
        self._httpd.shutdown()
        self.join(timeout=10)


def build_wsgi_application(app_root):
    openapi_engine = openapi_engines.OpenApiEngine(
        info=openapi_structures.OpenApiInfo(
            title="REST API Microservice",
            description="REST API Microservice for tests",
            version="1.2.3",
            contact=openapi_structures.OpenApiContact(
                name="Functional Tests",
                url="https://functional.tests/",
                email="functional@tests.local",
            ),
            license=openapi_structures.OpenApiApacheLicense(),
            terms_of_service="https://functional.tests/terms/",
        ),
        paths=openapi_structures.OpenApiPaths(),
        components=openapi_structures.OpenApiComponents(),
        security=[openapi_structures.OpenApiSecurity()],
        tags=openapi_structures.OpenApiTags(
            [
                openapi_structures.OpenApiTag(
                    name="functional-test",
                    description="Just functional tests",
                    external_docs=openapi_structures.OpenApiExternalDocs(
                        url="https://https://functional.tests/docs/",
                        description=("Functional tests external documentation"),
                    ),
                )
            ]
        ),
        external_docs=openapi_structures.OpenApiExternalDocs(
            url="https://https://functional.tests/docs/",
            description="Functional tests external documentation",
        ),
    )

    return errors_middleware.ErrorsHandlerMiddleware(
        retry_error_middleware.RetryOnErrorsMiddleware(
            middlewares.ContextMiddleware(
                application=applications.OpenApiApplication(
                    route_class=app_root,
                    openapi_engine=openapi_engine,
                ),
            ),
            exceptions=storage_exc.DeadLock,
            # set max_retry == 2 to speed up retry tests
            # execution
            max_retry=2,
        ),
    )


def main():
    rest_service = RESTService(
        bind_host="0.0.0.0",
        bind_port=8000,
        app_root=build_wsgi_application(app_root=routes.Root),
    )
    try:
        rest_service.run()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    engines.engine_factory.configure_factory(consts.get_database_uri())
    main()
